from typing import ( # noqa
    List,
    Dict,
    Optional,
    Tuple,
    Any,
    Union,
    Sequence,
    cast,
)

from chalice.config import Config, DeployedResources  # noqa
from chalice.deploy import models
from chalice.deploy.planner import Variable
from chalice.deploy.models import Instruction, StoreMultipleValue  # noqa

MarkedResource = Dict[str, List[models.RecordResource]]
ResourceValueType = Dict[str, Union[Sequence[Instruction], str]]
HandlerArgsType = List[Union[Dict[str, Any], str]]


class ResourceSweeper(object):

    specific_resources = (
        's3_event',
        'sns_event',
        'sqs_event',
        'kinesis_event',
        'dynamodb_event',
        'domain_name'
    )

    def __init__(self):
        # type: () -> None
        self.plan = models.Plan()
        self.marked = {}  # type: Dict

    def execute(self, plan, config):
        # type: (models.Plan, Config) -> None
        self.plan = plan
        self.marked = self._mark_resources()

        deployed = config.deployed_resources(config.chalice_stage)
        if deployed is not None:
            remaining = self._determine_remaining(deployed)
            self._plan_deletion(remaining, deployed)

    def _determine_sns_event(self, name, resource_values):
        # type: (str, Dict[str, str]) -> Optional[str]
        existing_topic = resource_values['topic']
        referenced_topic = [instruction for instruction in self.marked[name]
                            if instruction.name == 'topic' and
                            isinstance(instruction,
                                       models.RecordResourceValue)][0]
        if referenced_topic.value != existing_topic:
            return name
        return None

    def _determine_s3_event(self, name, resource_values):
        # type: (str, Dict[str, str]) -> Optional[str]
        # Special case, we have to check the resource values
        # to see if they've changed.  For s3 events, the resource
        # name is not tied to the bucket, which means if you change
        # the bucket, the resource name will stay the same.
        # So we match up the bucket referenced in the instruction
        # and the bucket recorded in the deployed values match up.
        # If they don't then we need to clean up the bucket config
        # referenced in the deployed values.
        bucket = [instruction for instruction in self.marked[name]
                  if instruction.name == 'bucket' and
                  isinstance(instruction,
                             models.RecordResourceValue)][0]
        if bucket.value != resource_values['bucket']:
            return name
        return None

    def _determine_sqs_event(self, name, resource_values):
        # type: (str, Dict[str, str]) -> Optional[str]
        existing_queue = resource_values['queue']
        referenced_queue = [instruction for instruction in self.marked[name]
                            if instruction.name == 'queue' and
                            isinstance(instruction,
                                       models.RecordResourceValue)][0]
        if referenced_queue.value != existing_queue:
            return name
        return None

    def _determine_kinesis_event(self, name, resource_values):
        # type: (str, Dict[str, str]) -> Optional[str]
        existing_stream = resource_values['stream']
        referenced_stream = [instruction for instruction in self.marked[name]
                             if instruction.name == 'stream' and
                             isinstance(instruction,
                                        models.RecordResourceValue)][0]
        if referenced_stream.value != existing_stream:
            return name
        return None

    def _determine_dynamodb_event(self, name, resource_values):
        # type: (str, Dict[str, str]) -> Optional[str]
        existing_stream_arn = resource_values['stream_arn']
        referenced_stream = [instruction for instruction in self.marked[name]
                             if instruction.name == 'stream_arn' and
                             isinstance(instruction,
                                        models.RecordResourceValue)][0]
        if referenced_stream.value != existing_stream_arn:
            return name
        return None

    def _determine_domain_name(self, name, resource_values):
        # type: (str, Dict[str, Any]) -> Optional[List[str]]
        api_mapping = resource_values.get('api_mapping')
        if not api_mapping:
            return None

        deployed_api_mappings_ids = {
            api_map['key']
            for api_map in api_mapping
        }
        api_mapping_data = (
            'rest_api_mapping',
            'websocket_api_mapping'
        )

        instructions = self.plan.instructions

        planned_api_mappings_ids = {
            instr.value[0]['key']
            for instr in instructions
            if isinstance(instr, StoreMultipleValue) and
            (instr.name in api_mapping_data and
                isinstance(instr.value[0], dict))
        }

        api_mappings_to_remove = list(
            deployed_api_mappings_ids - planned_api_mappings_ids
        )

        result_api_mappings = [
            "%s.api_mapping.%s" % (name, api_map)
            for api_map in api_mappings_to_remove
        ]
        return result_api_mappings

    def _determine_remaining(self, deployed):
        # type: (DeployedResources) -> List[str]
        remaining = []
        deployed_resource_names = reversed(deployed.resource_names())

        for name in deployed_resource_names:
            resource_values = deployed.resource_values(name)
            if name not in self.marked:
                remaining.append(name)
            elif resource_values['resource_type'] in self.specific_resources:
                method = '_determine_%s' % resource_values['resource_type']
                handler = getattr(self, method)
                resource_name = handler(name, resource_values)
                if resource_name:
                    if isinstance(resource_name, list):
                        remaining.extend(resource_name)
                    else:
                        remaining.append(resource_name)
        return remaining

    def _mark_resources(self):
        # type: () -> MarkedResource
        marked = {}  # type: MarkedResource
        for instruction in self.plan.instructions:
            if isinstance(instruction, models.RecordResource):
                marked.setdefault(instruction.resource_name, []).append(
                    instruction)
        return marked

    def _delete_domain_name(self,
                            resource_values  # type: Dict[str, Any]
                            ):
        # type: (...) -> ResourceValueType
        params = {
            'domain_name': resource_values['domain_name']
        }
        msg = 'Deleting custom domain name: %s\n' % resource_values['name']
        return {
            'instructions': (
                models.APICall(
                    method_name='delete_domain_name',
                    params=params,
                ),
            ),
            'message': msg
        }

    def _delete_api_mapping(self,
                            domain_name,   # type: str
                            api_mapping    # type: Dict[str, Any]
                            ):
        # type: (...) -> ResourceValueType
        if api_mapping['key'] == '/':
            path_key = '(none)'
        else:
            path_key = api_mapping['key'].lstrip("/")

        params = {
            'domain_name': domain_name,
            'path_key': path_key
        }
        msg = 'Deleting base path mapping from %s custom domain name: %s\n' % (
            domain_name, api_mapping['key']
        )
        return {
            'instructions': (
                models.APICall(
                    method_name='delete_api_mapping',
                    params=params,
                ),
            ),
            'message': msg
        }

    def _delete_lambda_function(self,
                                resource_values  # type: Dict[str, Any]
                                ):
        # type: (...) -> ResourceValueType
        msg = 'Deleting function: %s\n' % resource_values['lambda_arn']
        return {
            'instructions': (
                models.APICall(
                    method_name='delete_function',
                    params={'function_name': resource_values['lambda_arn']},
                ),
            ),
            'message': msg
        }

    def _delete_lambda_layer(self, resource_values):
        # type: (Dict[str, str]) -> ResourceValueType
        apicall = models.APICall(
            method_name='delete_layer_version',
            params={'layer_version_arn': resource_values[
                'layer_version_arn']})
        return {
            'instructions': (apicall,),
            'message': (
                "Deleting layer version: %s\n"
                % resource_values['layer_version_arn']
            )
        }

    def _delete_iam_role(self, resource_values):
        # type: (Dict[str, Any]) -> ResourceValueType
        return {
            'instructions': (
                models.APICall(
                    method_name='delete_role',
                    params={'name': resource_values['role_name']},
                ),
            ),
            'message': 'Deleting IAM role: %s\n' % resource_values['role_name']
        }

    def _delete_cloudwatch_event(self, resource_values):
        # type: (Dict[str, Any]) -> ResourceValueType
        return {
            'instructions': (
                models.APICall(
                    method_name='delete_rule',
                    params={'rule_name': resource_values['rule_name']},
                ),
            )
        }

    def _delete_rest_api(self, resource_values):
        # type: (Dict[str, Any]) -> ResourceValueType
        msg = 'Deleting Rest API: %s\n' % resource_values['rest_api_id']
        return {
            'instructions': (
                models.APICall(
                    method_name='delete_rest_api',
                    params={'rest_api_id': resource_values['rest_api_id']}
                ),
            ),
            'message': msg
        }

    def _delete_s3_event(self, resource_values):
        # type: (Dict[str, Any]) -> ResourceValueType
        bucket = resource_values['bucket']
        function_arn = resource_values['lambda_arn']
        return {
            'instructions': (
                models.BuiltinFunction('parse_arn', [function_arn],
                                       output_var='parsed_lambda_arn'),
                models.JPSearch('account_id', input_var='parsed_lambda_arn',
                                output_var='account_id'),
                models.APICall(
                    method_name='disconnect_s3_bucket_from_lambda',
                    params={'bucket': bucket, 'function_arn': function_arn}
                ),
                models.APICall(
                    method_name='remove_permission_for_s3_event',
                    params={'bucket': bucket, 'function_arn': function_arn,
                            'account_id': Variable('account_id')}
                ),
            )
        }

    def _delete_sns_event(self, resource_values):
        # type: (Dict[str, Any]) -> ResourceValueType
        subscription_arn = resource_values['subscription_arn']
        return {
            'instructions': (
                models.APICall(
                    method_name='unsubscribe_from_topic',
                    params={'subscription_arn': subscription_arn},
                ),
                models.APICall(
                    method_name='remove_permission_for_sns_topic',
                    params={
                        'topic_arn': resource_values['topic_arn'],
                        'function_arn': resource_values['lambda_arn'],
                    },
                ),
            )
        }

    def _delete_sqs_event(self, resource_values):
        # type: (Dict[str, Any]) -> ResourceValueType
        return {
            'instructions': (
                models.APICall(
                    method_name='remove_lambda_event_source',
                    params={'event_uuid': resource_values['event_uuid']},
                ),
            )
        }

    def _delete_kinesis_event(self, resource_values):
        # type: (Dict[str, Any]) -> ResourceValueType
        return {
            'instructions': (
                models.APICall(
                    method_name='remove_lambda_event_source',
                    params={'event_uuid': resource_values['event_uuid']},
                ),
            )
        }

    def _delete_dynamodb_event(self, resource_values):
        # type: (Dict[str, Any]) -> ResourceValueType
        return {
            'instructions': (
                models.APICall(
                    method_name='remove_lambda_event_source',
                    params={'event_uuid': resource_values['event_uuid']},
                ),
            )
        }

    def _delete_websocket_api(self, resource_values):
        # type: (Dict[str, Any]) -> ResourceValueType
        msg = 'Deleting Websocket API: %s\n' % (
            resource_values['websocket_api_id']
        )
        return {
            'instructions': (
                models.APICall(
                    method_name='delete_websocket_api',
                    params={'api_id': resource_values['websocket_api_id']},
                ),
            ),
            'message': msg
        }

    def _default_delete(self, resource_values):
        # type: (Dict[str, Any]) -> None
        err_msg = "Sweeper encountered an unknown resource: %s" % \
                  resource_values
        raise RuntimeError(err_msg)

    def _update_plan(self, instructions, message=None, insert=False):
        # type: (Tuple[Instruction], Optional[str], bool) -> None
        if insert:
            for instruction in instructions:
                self.plan.instructions.insert(
                    0, cast(Instruction, instruction)
                )
            if message:
                instr_id = id(self.plan.instructions[0])
                self.plan.messages[instr_id] = cast(
                    str, message
                )
        else:
            self.plan.instructions.extend(instructions)
            if message:
                self.plan.messages[id(self.plan.instructions[-1])] = message

    def _delete_domain_api_mappings(self, resource_values, name):
        # type: (Dict[str, Any], str) -> ResourceValueType
        path_key = name.split('.')[-1]

        api_mapping = {
            k: v
            for api_map in resource_values['api_mapping']
            for k, v in api_map.items()
            if api_map['key'] == path_key
        }  # type: Dict[str, str]

        resource_data = self._delete_api_mapping(
            resource_values['domain_name'],
            api_mapping
        )
        return resource_data

    def _plan_deletion(self,
                       remaining,  # type: List[str]
                       deployed,   # type: DeployedResources
                       ):
        # type: (...) -> None
        for name in remaining:
            resource_values = deployed.resource_values(name)

            resource_type = resource_values['resource_type']
            handler_args = [resource_values]  # type: HandlerArgsType
            insert = False
            if 'api_mapping' in name:
                resource_type = 'domain_api_mappings'
                handler_args.append(name)
                insert = True

            method_name = '_delete_%s' % resource_type
            handler = getattr(self, method_name, self._default_delete)
            resource_data = handler(*handler_args)
            instructions = cast(
                Tuple[Instruction],
                resource_data['instructions']
            )
            message = cast(Optional[str], resource_data.get('message'))
            self._update_plan(
                instructions,
                message,
                insert=insert
            )



"""Automatically reload chalice app when files change.

How It Works
============

This approach borrow from what django, flask, and other frameworks do.
Essentially, with reloading enabled ``chalice local`` will start up
a worker process that runs the dev http server.  This means there will
be a total of two processes running (both will show as ``chalice local``
in ps).  One process is the parent process.  It's job is to start up a child
process and restart it if it exits (due to a restart request).  The child
process is the process that actually starts up the web server for local mode.
The child process also sets up a watcher thread.  It's job is to monitor
directories for changes.  If a change is encountered it sys.exit()s the process
with a known RC (the RESTART_REQUEST_RC constant in the module).

The parent process runs in an infinite loop.  If the child process exits with
an RC of RESTART_REQUEST_RC the parent process starts up another child process.

The child worker is denoted by setting the ``CHALICE_WORKER`` env var.
If this env var is set, the process is intended to be a worker process (as
opposed the parent process which just watches for restart requests from the
worker process).

"""
import subprocess
import logging
import copy
import sys

from typing import MutableMapping, Type, Callable, Optional  # noqa

from chalice.cli.filewatch import RESTART_REQUEST_RC, WorkerProcess
from chalice.local import LocalDevServer, HTTPServerThread  # noqa


LOGGER = logging.getLogger(__name__)
WorkerProcType = Optional[Type[WorkerProcess]]


def get_best_worker_process():
    # type: () -> Type[WorkerProcess]
    try:
        from chalice.cli.filewatch.eventbased import WatchdogWorkerProcess
        LOGGER.debug("Using watchdog worker process.")
        return WatchdogWorkerProcess
    except ImportError:
        from chalice.cli.filewatch.stat import StatWorkerProcess
        LOGGER.debug("Using stat() based worker process.")
        return StatWorkerProcess


def start_parent_process(env):
    # type: (MutableMapping) -> None
    process = ParentProcess(env, subprocess.Popen)
    process.main()


def start_worker_process(server_factory, root_dir, worker_process_cls=None):
    # type: (Callable[[], LocalDevServer], str, WorkerProcType) -> int
    if worker_process_cls is None:
        worker_process_cls = get_best_worker_process()
    t = HTTPServerThread(server_factory)
    worker = worker_process_cls(t)
    LOGGER.debug("Starting worker...")
    rc = worker.main(root_dir)
    LOGGER.info("Restarting local dev server.")
    return rc


class ParentProcess(object):
    """Spawns a child process and restarts it as needed."""
    def __init__(self, env, popen):
        # type: (MutableMapping, Type[subprocess.Popen]) -> None
        self._env = copy.copy(env)
        self._popen = popen

    def main(self):
        # type: () -> None
        # This method launches a child worker and restarts it if it
        # exits with RESTART_REQUEST_RC.  This method doesn't return.
        # A user can Ctrl-C to stop the parent process.
        while True:
            self._env['CHALICE_WORKER'] = 'true'
            LOGGER.debug("Parent process starting child worker process...")
            process = self._popen(sys.argv, env=self._env)
            try:
                process.communicate()
                if process.returncode != RESTART_REQUEST_RC:
                    return
            except KeyboardInterrupt:
                process.terminate()
                raise


def run_with_reloader(server_factory, env, root_dir, worker_process_cls=None):
    # type: (Callable, MutableMapping, str, WorkerProcType) -> int
    # This function is invoked in two possible modes, as the parent process
    # or as a chalice worker.
    try:
        if env.get('CHALICE_WORKER') is not None:
            # This is a chalice worker.  We need to start the main dev server
            # in a daemon thread and install a file watcher.
            return start_worker_process(server_factory, root_dir,
                                        worker_process_cls)
        else:
            # This is the parent process.  It's just is to spawn an identical
            # process but with the ``CHALICE_WORKER`` env var set.  It then
            # will monitor this process and restart it if it exits with a
            # RESTART_REQUEST exit code.
            start_parent_process(env)
    except KeyboardInterrupt:
        pass
    return 0



# pylint: disable=too-many-lines
import re
import json
from collections import OrderedDict

from typing import List, Dict, Any, Optional, Union, Tuple, Set, cast  # noqa
from typing import Sequence  # noqa

from chalice.config import Config, DeployedResources  # noqa
from chalice.utils import OSUtils  # noqa
from chalice.deploy import models
from chalice.awsclient import TypedAWSClient, ResourceDoesNotExistError  # noqa


InstructionMsg = Union[models.Instruction, Tuple[models.Instruction, str]]
MarkedResource = Dict[str, List[models.RecordResource]]
CacheTuples = Union[Tuple[str, str, str], Tuple[str, str]]
ApiMap = Union[models.RestAPI, models.WebsocketAPI]


class RemoteState(object):
    def __init__(self, client, deployed_resources):
        # type: (TypedAWSClient, DeployedResources) -> None
        self._client = client
        self._cache = {}  # type: Dict[CacheTuples, bool]
        self._deployed_resources = deployed_resources

    def _cache_key(self, resource):
        # type: (models.ManagedModel) -> CacheTuples
        if isinstance(resource, models.APIMapping):
            return (
                resource.resource_type,
                resource.resource_name,
                resource.mount_path
            )
        return resource.resource_type, resource.resource_name

    def resource_deployed_values(self, resource):
        # type: (models.ManagedModel) -> Dict[str, Any]
        try:
            return self._deployed_resources.resource_values(
                resource.resource_name)
        except ValueError:
            return self._dynamically_lookup_values(resource)

    def _dynamically_lookup_values(self, resource):
        # type: (models.ManagedModel) -> Dict[str, str]
        if isinstance(resource, models.ManagedIAMRole):
            arn = self._client.get_role_arn_for_name(resource.role_name)
            return {
                "role_name": resource.role_name,
                "role_arn": arn,
                "name": resource.resource_name,
                "resource_type": "iam_role",
            }
        raise ValueError("Deployed values for resource does not exist: %s"
                         % resource.resource_name)

    def resource_exists(self, resource, *args):
        # type: (models.ManagedModel, Optional[Any]) -> bool
        key = self._cache_key(resource)
        if key in self._cache:
            return self._cache[key]
        try:
            handler = getattr(self, '_resource_exists_%s'
                              % resource.__class__.__name__.lower())
        except AttributeError:
            raise ValueError("RemoteState received an unsupported resource: %s"
                             % resource.resource_type)
        result = handler(resource, *args)
        self._cache[key] = result
        return result

    def _resource_exists_snslambdasubscription(self, resource):
        # type: (models.SNSLambdaSubscription) -> bool
        try:
            deployed_values = self._deployed_resources.resource_values(
                resource.resource_name)
        except ValueError:
            return False
        return self._client.verify_sns_subscription_current(
            deployed_values['subscription_arn'],
            topic_name=resource.topic,
            function_arn=deployed_values['lambda_arn'],
        )

    def _resource_exists_sqseventsource(self, resource):
        # type: (models.SQSEventSource) -> bool
        try:
            deployed_values = self._deployed_resources.resource_values(
                resource.resource_name)
        except ValueError:
            return False
        if isinstance(resource.queue, models.QueueARN):
            resource_name = resource.queue.queue_name
        else:
            resource_name = resource.queue
        return self._client.verify_event_source_current(
            event_uuid=deployed_values['event_uuid'],
            resource_name=resource_name,
            service_name='sqs',
            function_arn=deployed_values['lambda_arn'],
        )

    def _resource_exists_kinesiseventsource(self, resource):
        # type: (models.KinesisEventSource) -> bool
        try:
            deployed_values = self._deployed_resources.resource_values(
                resource.resource_name)
        except ValueError:
            return False
        return self._client.verify_event_source_current(
            event_uuid=deployed_values['event_uuid'],
            resource_name='stream/%s' % resource.stream,
            service_name='kinesis',
            function_arn=deployed_values['lambda_arn'],
        )

    def _resource_exists_dynamodbeventsource(self, resource):
        # type: (models.DynamoDBEventSource) -> bool
        try:
            deployed_values = self._deployed_resources.resource_values(
                resource.resource_name)
        except ValueError:
            return False
        return self._client.verify_event_source_arn_current(
            event_uuid=deployed_values['event_uuid'],
            event_source_arn=deployed_values['stream_arn'],
            function_arn=deployed_values['lambda_arn'],
        )

    def _resource_exists_lambdalayer(self, resource):
        # type: (models.LambdaLayer) -> bool
        try:
            deployed_values = self._deployed_resources.resource_values(
                resource.resource_name)
        except ValueError:
            return False
        return bool(self._client.get_layer_version(
            deployed_values['layer_version_arn']))

    def _resource_exists_lambdafunction(self, resource):
        # type: (models.LambdaFunction) -> bool
        return self._client.lambda_function_exists(resource.function_name)

    def _resource_exists_managediamrole(self, resource):
        # type: (models.ManagedIAMRole) -> bool
        try:
            self._client.get_role_arn_for_name(resource.role_name)
            return True
        except ResourceDoesNotExistError:
            return False

    def _resource_exists_apimapping(self, resource, domain_name):
        # type: (models.APIMapping, str) -> bool
        map_key = resource.mount_path
        if map_key == '(none)':
            map_key = ''
        elif map_key.startswith('/'):
            map_key = map_key.lstrip('/')

        return self._client.api_mapping_exists(domain_name, map_key)

    def _resource_exists_domainname(self, resource):
        # type: (models.DomainName) -> bool
        if resource.protocol == models.APIType.WEBSOCKET:
            return self._client.domain_name_exists_v2(
                resource.domain_name)
        return self._client.domain_name_exists(resource.domain_name)

    def _resource_exists_restapi(self, resource):
        # type: (models.RestAPI) -> bool
        try:
            deployed_values = self._deployed_resources.resource_values(
                resource.resource_name)
        except ValueError:
            return False
        rest_api_id = deployed_values['rest_api_id']
        return bool(self._client.get_rest_api(rest_api_id))

    def _resource_exists_websocketapi(self, resource):
        # type: (models.WebsocketAPI) -> bool
        try:
            deployed_values = self._deployed_resources.resource_values(
                resource.resource_name)
        except ValueError:
            return False
        api_id = deployed_values['websocket_api_id']
        return self._client.websocket_api_exists(api_id)


class PlanStage(object):
    def __init__(self, remote_state, osutils):
        # type: (RemoteState, OSUtils) -> None
        self._remote_state = remote_state
        self._osutils = osutils

    def execute(self, resources):
        # type: (List[models.Model]) -> models.Plan
        plan = []  # type: List[models.Instruction]
        messages = {}  # type: Dict[int, str]
        for resource in resources:
            name = '_plan_%s' % resource.__class__.__name__.lower()
            handler = getattr(self, name, None)
            if handler is not None:
                result = handler(resource)
                if result:
                    self._add_result_to_plan(result, plan, messages)
        return models.Plan(plan, messages)

    def _add_result_to_plan(self,
                            result,    # type: Sequence[InstructionMsg]
                            plan,      # type: List[models.Instruction]
                            messages,  # type: Dict[int, str]
                            ):
        # type: (...) -> None
        for single in result:
            if isinstance(single, tuple):
                instruction, message = single
                plan.append(instruction)
                messages[id(instruction)] = message
            else:
                plan.append(single)

    # TODO: This code will likely be refactored and pulled into
    # per-resource classes so the PlanStage object doesn't need
    # to know about every type of resource.

    def _add_apimapping_plan(self,
                             resource,    # type: models.APIMapping
                             domain_name  # type: models.DomainName
                             ):
        # type: (...) -> Sequence[InstructionMsg]
        api_calls = []  # type: List[InstructionMsg]
        params = {
            'domain_name': domain_name.domain_name,
            'path_key': resource.mount_path,
            'stage': resource.api_gateway_stage
        }  # type: Dict[str, Any]
        if domain_name.protocol == models.APIType.WEBSOCKET:
            params['api_id'] = Variable('websocket_api_id')
            variable_name = 'websocket_api_mapping'
            api_call = models.APICall(
                method_name='create_api_mapping',
                params=params,
                output_var='api_mapping'
            )
        else:
            params['api_id'] = Variable('rest_api_id')
            variable_name = 'rest_api_mapping'
            api_call = models.APICall(
                method_name='create_base_path_mapping',
                params=params,
                output_var='api_mapping'
            )

        if not self._remote_state.resource_exists(
                resource, domain_name.domain_name
        ):
            path_to_print = '/'
            if resource.mount_path != '(none)' and \
                    not resource.mount_path.startswith("/"):
                path_to_print = '/%s' % resource.mount_path
            api_calls.extend([
                (api_call, "Creating api mapping: %s\n" % path_to_print),
                models.StoreMultipleValue(
                    name=variable_name,
                    value=[Variable('api_mapping')]
                ),
                models.RecordResourceVariable(
                    resource_type='domain_name',
                    resource_name=domain_name.resource_name,
                    name='api_mapping',
                    variable_name=variable_name
                ),
            ])
        else:
            deployed = self._remote_state.resource_deployed_values(
                domain_name
            )
            for api_mapping in deployed['api_mapping']:

                mount_path = api_mapping['key'].lstrip('/')
                if not mount_path:
                    mount_path = '(none)'
                if mount_path != resource.mount_path:
                    continue

                api_calls.extend([
                    models.StoreMultipleValue(
                        name=variable_name,
                        value=[api_mapping]
                    ),
                    models.RecordResourceVariable(
                        resource_type='domain_name',
                        resource_name=domain_name.resource_name,
                        name='api_mapping',
                        variable_name=variable_name
                    ),
                ])
        return api_calls

    def _add_domainname_plan(self, resource, endpoint_type):
        # type: (models.DomainName, str) -> Sequence[InstructionMsg]
        api_calls = []  # type: List[InstructionMsg]

        params = {
            'protocol': resource.protocol.value,
            'tags': resource.tags,
            'endpoint_type': endpoint_type,
            'domain_name': resource.domain_name,
        }
        params['certificate_arn'] = resource.certificate_arn
        if resource.tls_version is not None:
            params['security_policy'] = resource.tls_version.value

        if not self._remote_state.resource_exists(resource):
            domain_name_api_call = (
                models.APICall(
                    method_name='create_domain_name',
                    params=params,
                    output_var=resource.resource_name
                ),
                "Creating custom domain name: %s\n" % resource.domain_name
            )

        else:
            domain_name_api_call = (
                models.APICall(
                    method_name='update_domain_name',
                    params=params,
                    output_var=resource.resource_name
                ),
                "Updating custom domain name: %s\n" % resource.domain_name
            )

        api_calls.extend([
            domain_name_api_call,
            models.StoreValue(
                name='hosted_zone_id',
                value=KeyDataVariable(resource.resource_name,
                                      'hosted_zone_id')
            ),
            models.RecordResourceVariable(
                resource_type='domain_name',
                resource_name=resource.resource_name,
                name='hosted_zone_id',
                variable_name='hosted_zone_id'
            ),
            models.StoreValue(
                name='alias_domain_name',
                value=KeyDataVariable(resource.resource_name,
                                      'alias_domain_name')
            ),
            models.RecordResourceVariable(
                resource_type='domain_name',
                resource_name=resource.resource_name,
                name='alias_domain_name',
                variable_name='alias_domain_name'
            ),
            models.StoreValue(
                name='certificate_arn',
                value=KeyDataVariable(resource.resource_name,
                                      'certificate_arn')
            ),
            models.RecordResourceVariable(
                resource_type='domain_name',
                resource_name=resource.resource_name,
                name='certificate_arn',
                variable_name='certificate_arn'
            ),
            models.StoreValue(
                name='security_policy',
                value=KeyDataVariable(resource.resource_name,
                                      'security_policy')
            ),
            models.RecordResourceVariable(
                resource_type='domain_name',
                resource_name=resource.resource_name,
                name='security_policy',
                variable_name='security_policy'
            ),
            models.RecordResourceValue(
                resource_type='domain_name',
                resource_name=resource.resource_name,
                name='domain_name',
                value=resource.domain_name
            )
        ])
        return api_calls

    def _plan_lambdalayer(self, resource):
        # type: (models.LambdaLayer) -> Sequence[InstructionMsg]

        api_calls = []  # type: List[InstructionMsg]
        filename = cast(str, resource.deployment_package.filename)

        # Automatically clean up old layer versions.
        # See:
        # https://docs.aws.amazon.com/lambda/latest/dg/API_DeleteLayerVersion.html
        msg = 'Creating'
        if self._remote_state.resource_exists(resource):
            state = self._remote_state.resource_deployed_values(resource)
            # Deleting a layer version won't break functions still using it.
            # From the doc link above:
            #
            # "To avoid breaking functions, a copy of the version remains in
            # Lambda until no functions refer to it."
            api_calls.append(
                models.APICall(
                    method_name='delete_layer_version',
                    params={'layer_version_arn': state['layer_version_arn']}
                )
            )
            msg = 'Updating'

        api_calls.extend([(
            models.APICall(
                method_name='publish_layer',
                params={'layer_name': resource.layer_name,
                        'zip_contents': self._osutils.get_file_contents(
                            filename, binary=True),
                        'runtime': resource.runtime},
                output_var='layer_version_arn'
            ), "%s lambda layer: %s\n" % (msg, resource.layer_name)),
            models.RecordResourceVariable(
                resource_type='lambda_layer',
                resource_name=resource.resource_name,
                name='layer_version_arn',
                variable_name='layer_version_arn',
        )])
        return api_calls

    def _plan_lambdafunction(self, resource):
        # type: (models.LambdaFunction) -> Sequence[InstructionMsg]
        role_arn = self._get_role_arn(resource.role)
        # Make mypy happy, it complains if we don't "declare" this upfront.
        params = {}  # type: Dict[str, Any]
        varname = '%s_lambda_arn' % resource.resource_name
        # Not sure the best way to express this via mypy, but we know
        # that in the build stage we replace the deployment package
        # name with the actual filename generated from the pip
        # packager.  For now we resort to a cast.
        filename = cast(str, resource.deployment_package.filename)

        if resource.reserved_concurrency is None:
            concurrency_api_call = models.APICall(
                method_name='delete_function_concurrency',
                params={
                    'function_name': resource.function_name,
                },
                output_var='reserved_concurrency_result'
            )
        else:
            concurrency = resource.reserved_concurrency
            concurrency_api_call = (
                models.APICall(
                    method_name='put_function_concurrency',
                    params={
                        'function_name': resource.function_name,
                        'reserved_concurrent_executions': concurrency,
                    },
                    output_var='reserved_concurrency_result'),
                "Updating lambda function concurrency limit: %s\n"
                % resource.function_name
            )

        api_calls = []  # type: List[InstructionMsg]
        layers = []  # type: List[Any]
        if resource.managed_layer is not None:
            layers.append(Variable('layer_version_arn'))
        if resource.layers:
            layers.extend(resource.layers)

        if not self._remote_state.resource_exists(resource):
            params = {
                'function_name': resource.function_name,
                'role_arn': role_arn,
                'zip_contents': self._osutils.get_file_contents(
                    filename, binary=True),
                'runtime': resource.runtime,
                'handler': resource.handler,
                'environment_variables': resource.environment_variables,
                'xray': resource.xray,
                'tags': resource.tags,
                'timeout': resource.timeout,
                'memory_size': resource.memory_size,
                'security_group_ids': resource.security_group_ids,
                'subnet_ids': resource.subnet_ids,
                'layers': layers
            }

            api_calls.extend([
                (models.APICall(
                    method_name='create_function',
                    params=params,
                    output_var=varname,
                ), "Creating lambda function: %s\n" % resource.function_name),
                models.RecordResourceVariable(
                    resource_type='lambda_function',
                    resource_name=resource.resource_name,
                    name='lambda_arn',
                    variable_name=varname,
                )
            ])
        else:
            # TODO: Consider a smarter diff where we check if we even need
            # to do an update() API call.
            params = {
                'function_name': resource.function_name,
                'role_arn': role_arn,
                'zip_contents': self._osutils.get_file_contents(
                    filename, binary=True),
                'runtime': resource.runtime,
                'environment_variables': resource.environment_variables,
                'xray': resource.xray,
                'tags': resource.tags,
                'timeout': resource.timeout,
                'memory_size': resource.memory_size,
                'security_group_ids': resource.security_group_ids,
                'subnet_ids': resource.subnet_ids,
                'layers': layers
            }
            api_calls.extend([
                (models.APICall(
                    method_name='update_function',
                    params=params,
                    output_var='update_function_result',
                ), "Updating lambda function: %s\n" % resource.function_name),
                models.JPSearch(
                    'FunctionArn',
                    input_var='update_function_result',
                    output_var=varname,
                ),
                models.RecordResourceVariable(
                    resource_type='lambda_function',
                    resource_name=resource.resource_name,
                    name='lambda_arn',
                    variable_name=varname,
                )
            ])
        api_calls.append(concurrency_api_call)
        return api_calls

    def _plan_managediamrole(self, resource):
        # type: (models.ManagedIAMRole) -> Sequence[InstructionMsg]
        document = resource.policy.document
        role_exists = self._remote_state.resource_exists(resource)
        varname = '%s_role_arn' % resource.role_name
        if not role_exists:
            return [
                models.BuiltinFunction(
                    'service_principal',
                    ['lambda'],
                    output_var='lambda_service_principal',
                ),
                models.JPSearch('principal',
                                input_var='lambda_service_principal',
                                output_var='lambda_principal'),
                models.StoreValue(
                    name='lambda_principal',
                    value=StringFormat('{lambda_principal}',
                                       ['lambda_principal']),
                ),
                models.StoreValue(
                    name='lambda_trust_policy',
                    value={
                        "Version": "2012-10-17",
                        "Statement": [{
                            "Sid": "",
                            "Effect": "Allow",
                            "Principal": {
                                "Service": Variable('lambda_principal')
                            },
                            "Action": "sts:AssumeRole"
                        }]
                    },
                ),
                (models.APICall(
                    method_name='create_role',
                    params={'name': resource.role_name,
                            'trust_policy': Variable('lambda_trust_policy'),
                            'policy': document},
                    output_var=varname,
                ), "Creating IAM role: %s\n" % resource.role_name),
                models.RecordResourceVariable(
                    resource_type='iam_role',
                    resource_name=resource.resource_name,
                    name='role_arn',
                    variable_name=varname,
                ),
                models.RecordResourceValue(
                    resource_type='iam_role',
                    resource_name=resource.resource_name,
                    name='role_name',
                    value=resource.role_name,
                )
            ]
        role_arn = self._remote_state.resource_deployed_values(
            resource)['role_arn']
        return [
            models.StoreValue(name=varname, value=role_arn),
            (models.APICall(
                method_name='put_role_policy',
                params={'role_name': resource.role_name,
                        'policy_name': resource.role_name,
                        'policy_document': document},
            ), "Updating policy for IAM role: %s\n" % resource.role_name),
            models.RecordResourceVariable(
                resource_type='iam_role',
                resource_name=resource.resource_name,
                name='role_arn',
                variable_name=varname,
            ),
            models.RecordResourceValue(
                resource_type='iam_role',
                resource_name=resource.resource_name,
                name='role_name',
                value=resource.role_name,
            )
        ]

    def _plan_snslambdasubscription(self, resource):
        # type: (models.SNSLambdaSubscription) -> Sequence[InstructionMsg]
        function_arn = Variable(
            '%s_lambda_arn' % resource.lambda_function.resource_name
        )
        topic_arn_varname = '%s_topic_arn' % resource.resource_name
        subscribe_varname = '%s_subscription_arn' % resource.resource_name

        instruction_for_topic_arn = []  # type: List[InstructionMsg]
        if re.match(r"^arn:aws[a-z\-]*:sns:", resource.topic):
            instruction_for_topic_arn += [
                models.StoreValue(
                    name=topic_arn_varname,
                    value=resource.topic,
                )
            ]
        else:
            # To keep the user API simple, we only require the topic
            # name and not the ARN.  However, the APIs require the topic
            # ARN so we need to reconstruct it here in the planner.
            instruction_for_topic_arn += self._arn_parse_instructions(
                function_arn) + [
                models.StoreValue(
                    name=topic_arn_varname,
                    value=StringFormat(
                        'arn:{partition}:sns:{region_name}:{account_id}:%s' % (
                            resource.topic
                        ),
                        ['partition', 'region_name', 'account_id'],
                    ),
                )
            ]
        if self._remote_state.resource_exists(resource):
            # Given there's nothing about an SNS subscription you can
            # configure for now, if the resource exists, we don't do
            # anything.  The resource sweeper will verify that if the
            # subscription doesn't actually apply that we should unsubscribe
            # from the topic.
            deployed = self._remote_state.resource_deployed_values(resource)
            subscription_arn = deployed['subscription_arn']
            return instruction_for_topic_arn + self._batch_record_resource(
                'sns_event', resource.resource_name, {
                    'topic': resource.topic,
                    'lambda_arn': Variable(function_arn.name),
                    'subscription_arn': subscription_arn,
                    'topic_arn': Variable(topic_arn_varname),
                }
            )
        return instruction_for_topic_arn + [
            models.APICall(
                method_name='add_permission_for_sns_topic',
                params={'topic_arn': Variable(topic_arn_varname),
                        'function_arn': function_arn},
            ),
            (models.APICall(
                method_name='subscribe_function_to_topic',
                params={'topic_arn': Variable(topic_arn_varname),
                        'function_arn': function_arn},
                output_var=subscribe_varname,
            ), 'Subscribing %s to SNS topic %s\n'
                % (resource.lambda_function.function_name, resource.topic)
            )
        ] + self._batch_record_resource(
            'sns_event', resource.resource_name, {
                'topic': resource.topic,
                'lambda_arn': Variable(function_arn.name),
                'subscription_arn': Variable(subscribe_varname),
                'topic_arn': Variable(topic_arn_varname),
            }
        )

    def _plan_sqseventsource(self, resource):
        # type: (models.SQSEventSource) -> Sequence[InstructionMsg]
        queue_arn_varname = '%s_queue_arn' % resource.resource_name
        uuid_varname = '%s_uuid' % resource.resource_name
        function_arn = Variable(
            '%s_lambda_arn' % resource.lambda_function.resource_name
        )
        if not isinstance(resource.queue, models.QueueARN):
            instruction_for_queue_arn = self._arn_parse_instructions(
                function_arn)
            instruction_for_queue_arn.append(
                models.StoreValue(
                    name=queue_arn_varname,
                    value=StringFormat(
                        'arn:{partition}:sqs:{region_name}:{account_id}:%s' % (
                            resource.queue
                        ),
                        ['partition', 'region_name', 'account_id'],
                    ),
                )
            )
            queue_name = resource.queue
        else:
            instruction_for_queue_arn = [
                models.StoreValue(
                    name=queue_arn_varname,
                    value=resource.queue.arn,
                )
            ]
            queue_name = resource.queue.queue_name
        if self._remote_state.resource_exists(resource):
            deployed = self._remote_state.resource_deployed_values(resource)
            uuid = deployed['event_uuid']
            return instruction_for_queue_arn + [
                models.APICall(
                    method_name='update_lambda_event_source',
                    params={'event_uuid': uuid,
                            'batch_size': resource.batch_size,
                            'maximum_batching_window_in_seconds':
                                resource.maximum_batching_window_in_seconds}
                )
            ] + self._batch_record_resource(
                'sqs_event', resource.resource_name, {
                    'queue_arn': deployed['queue_arn'],
                    'event_uuid': uuid,
                    'queue': queue_name,
                    'lambda_arn': deployed['lambda_arn'],
                }
            )
        return instruction_for_queue_arn + [
            (models.APICall(
                method_name='create_lambda_event_source',
                params={'event_source_arn': Variable(queue_arn_varname),
                        'batch_size': resource.batch_size,
                        'maximum_batching_window_in_seconds':
                            resource.maximum_batching_window_in_seconds,
                        'function_name': function_arn},
                output_var=uuid_varname,
            ), 'Subscribing %s to SQS queue %s\n'
                % (resource.lambda_function.function_name, resource.queue)
            ),
        ] + self._batch_record_resource(
            'sqs_event', resource.resource_name, {
                'queue_arn': Variable(queue_arn_varname),
                'event_uuid': Variable(uuid_varname),
                'queue': queue_name,
                'lambda_arn': Variable(function_arn.name)
            }
        )

    def _plan_kinesiseventsource(self, resource):
        # type: (models.KinesisEventSource) -> Sequence[InstructionMsg]
        stream_arn_varname = '%s_stream_arn' % resource.resource_name
        uuid_varname = '%s_uuid' % resource.resource_name
        function_arn = Variable(
            '%s_lambda_arn' % resource.lambda_function.resource_name
        )
        instruction_for_stream_arn = self._arn_parse_instructions(function_arn)
        instruction_for_stream_arn.append(
            models.StoreValue(
                name=stream_arn_varname,
                value=StringFormat(
                    'arn:{partition}:kinesis:{region_name}:{account_id}:'
                    'stream/%s' % resource.stream,
                    ['partition', 'region_name', 'account_id'],
                ),
            )
        )
        if self._remote_state.resource_exists(resource):
            deployed = self._remote_state.resource_deployed_values(resource)
            uuid = deployed['event_uuid']
            return instruction_for_stream_arn + [
                models.APICall(
                    method_name='update_lambda_event_source',
                    params={'event_uuid': uuid,
                            'batch_size': resource.batch_size,
                            'maximum_batching_window_in_seconds':
                                resource.maximum_batching_window_in_seconds}
                )
            ] + self._batch_record_resource(
                'kinesis_event', resource.resource_name, {
                    'kinesis_arn': deployed['kinesis_arn'],
                    'event_uuid': uuid,
                    'stream': resource.stream,
                    'lambda_arn': deployed['lambda_arn'],
                }
            )
        return instruction_for_stream_arn + [
            (models.APICall(
                method_name='create_lambda_event_source',
                params={'event_source_arn': Variable(stream_arn_varname),
                        'batch_size': resource.batch_size,
                        'function_name': function_arn,
                        'starting_position': resource.starting_position,
                        'maximum_batching_window_in_seconds':
                            resource.maximum_batching_window_in_seconds},
                output_var=uuid_varname,
            ), 'Subscribing %s to Kinesis stream %s\n'
                % (resource.lambda_function.function_name, resource.stream)
            )
        ] + self._batch_record_resource(
            'kinesis_event', resource.resource_name, {
                'kinesis_arn': Variable(stream_arn_varname),
                'event_uuid': Variable(uuid_varname),
                'stream': resource.stream,
                'lambda_arn': Variable(function_arn.name),
            }
        )

    def _plan_dynamodbeventsource(self, resource):
        # type: (models.DynamoDBEventSource) -> Sequence[InstructionMsg]
        uuid_varname = '%s_uuid' % resource.resource_name
        function_arn = Variable(
            '%s_lambda_arn' % resource.lambda_function.resource_name
        )
        instructions = []  # type: List[InstructionMsg]
        if self._remote_state.resource_exists(resource):
            deployed = self._remote_state.resource_deployed_values(resource)
            uuid = deployed['event_uuid']
            return instructions + [
                models.APICall(
                    method_name='update_lambda_event_source',
                    params={'event_uuid': uuid,
                            'batch_size': resource.batch_size,
                            'maximum_batching_window_in_seconds':
                                resource.maximum_batching_window_in_seconds}
                )
            ] + self._batch_record_resource(
                'dynamodb_event', resource.resource_name, {
                    'stream_arn': deployed['stream_arn'],
                    'event_uuid': deployed['event_uuid'],
                    'lambda_arn': deployed['lambda_arn'],
                }
            )
        return instructions + [
            (models.APICall(
                method_name='create_lambda_event_source',
                params={'event_source_arn': resource.stream_arn,
                        'batch_size': resource.batch_size,
                        'function_name': function_arn,
                        'starting_position': resource.starting_position,
                        'maximum_batching_window_in_seconds':
                            resource.maximum_batching_window_in_seconds},
                output_var=uuid_varname,
            ), 'Subscribing %s to DynamoDB stream %s\n'
                % (resource.lambda_function.function_name,
                   resource.stream_arn))
        ] + self._batch_record_resource(
            'dynamodb_event', resource.resource_name, {
                'stream_arn': resource.stream_arn,
                'event_uuid': Variable(uuid_varname),
                'lambda_arn': function_arn,
            }
        )

    def _arn_parse_instructions(self, function_arn):
        # type: (Variable) -> List[InstructionMsg]
        instruction_for_stream_arn = [
            models.BuiltinFunction('parse_arn', [function_arn],
                                   output_var='parsed_lambda_arn'),
            models.JPSearch('account_id', input_var='parsed_lambda_arn',
                            output_var='account_id'),
            models.JPSearch('region', input_var='parsed_lambda_arn',
                            output_var='region_name'),
            models.JPSearch('partition', input_var='parsed_lambda_arn',
                            output_var='partition'),
        ]  # type: List[InstructionMsg]
        return instruction_for_stream_arn

    def _plan_s3bucketnotification(self, resource):
        # type: (models.S3BucketNotification) -> Sequence[InstructionMsg]
        function_arn = Variable(
            '%s_lambda_arn' % resource.lambda_function.resource_name
        )
        return self._arn_parse_instructions(function_arn) + [
            models.APICall(
                method_name='add_permission_for_s3_event',
                params={'bucket': resource.bucket,
                        'function_arn': function_arn,
                        'account_id': Variable('account_id')},
            ),
            (models.APICall(
                method_name='connect_s3_bucket_to_lambda',
                params={'bucket': resource.bucket,
                        'function_arn': function_arn,
                        'prefix': resource.prefix,
                        'suffix': resource.suffix,
                        'events': resource.events}
            ), 'Configuring S3 events in bucket %s to function %s\n'
                % (resource.bucket, resource.lambda_function.function_name)
            ),
            models.RecordResourceValue(
                resource_type='s3_event',
                resource_name=resource.resource_name,
                name='bucket',
                value=resource.bucket,
            ),
            models.RecordResourceVariable(
                resource_type='s3_event',
                resource_name=resource.resource_name,
                name='lambda_arn',
                variable_name=function_arn.name,
            ),
        ]

    def _create_cloudwatchevent(self, resource):
        # type: (models.CloudWatchEventBase) -> Sequence[InstructionMsg]

        function_arn = Variable(
            '%s_lambda_arn' % resource.lambda_function.resource_name
        )

        params = {'rule_name': resource.rule_name}
        if isinstance(resource, models.ScheduledEvent):
            resource = cast(models.ScheduledEvent, resource)
            params['schedule_expression'] = resource.schedule_expression
            if resource.rule_description is not None:
                params['rule_description'] = resource.rule_description
        else:
            resource = cast(models.CloudWatchEvent, resource)
            params['event_pattern'] = resource.event_pattern

        plan = [
            models.APICall(
                method_name='get_or_create_rule_arn',
                params=params,
                output_var='rule-arn',
            ),
            models.APICall(
                method_name='connect_rule_to_lambda',
                params={'rule_name': resource.rule_name,
                        'function_arn': function_arn}
            ),
            models.APICall(
                method_name='add_permission_for_cloudwatch_event',
                params={'rule_arn': Variable('rule-arn'),
                        'function_arn': function_arn},
            ),
            # You need to remove targets (which have IDs)
            # before you can delete a rule.
            models.RecordResourceValue(
                resource_type='cloudwatch_event',
                resource_name=resource.resource_name,
                name='rule_name',
                value=resource.rule_name,
            )
        ]
        return plan

    def _plan_cloudwatchevent(self, resource):
        # type: (models.CloudWatchEvent) -> Sequence[InstructionMsg]
        return self._create_cloudwatchevent(resource)

    def _plan_scheduledevent(self, resource):
        # type: (models.ScheduledEvent) -> Sequence[InstructionMsg]
        return self._create_cloudwatchevent(resource)

    def _create_websocket_function_configs(self, resource):
        # type: (models.WebsocketAPI) -> Dict[str, Dict[str, Any]]
        configs = OrderedDict()  # type: Dict[str, Dict[str, Any]]
        if resource.connect_function is not None:
            configs['connect'] = self._create_websocket_function_config(
                resource.connect_function)
        if resource.message_function is not None:
            configs['message'] = self._create_websocket_function_config(
                resource.message_function)
        if resource.disconnect_function is not None:
            configs['disconnect'] = self._create_websocket_function_config(
                resource.disconnect_function)
        return configs

    def _create_websocket_function_config(self, function):
        # type: (models.LambdaFunction) -> Dict[str, Any]
        varname = '%s_lambda_arn' % function.resource_name
        return {
            'function': function,
            'name': function.function_name,
            'varname': varname,
            'lambda_arn_var': Variable(varname),
        }

    def _inject_websocket_integrations(self, configs):
        # type: (Dict[str, Any]) -> Sequence[InstructionMsg]
        instructions = []  # type: List[InstructionMsg]
        for key, config in configs.items():
            instructions.append(
                models.StoreValue(
                    name='websocket-%s-integration-lambda-path' % key,
                    value=StringFormat(
                        'arn:{partition}:apigateway:{region_name}:lambda:path/'
                        '2015-03-31/functions/arn:{partition}'
                        ':lambda:{region_name}:{account_id}:function'
                        ':%s/invocations' % config['name'],
                        ['partition', 'region_name', 'account_id'],
                    ),
                ),
            )
            instructions.append(
                models.APICall(
                    method_name='create_websocket_integration',
                    params={
                        'api_id': Variable('websocket_api_id'),
                        'lambda_function': Variable(
                            'websocket-%s-integration-lambda-path' % key),
                        'handler_type': key,
                    },
                    output_var='%s-integration-id' % key,
                ),
            )
        return instructions

    def _create_route_for_key(self, route_key):
        # type: (str) -> models.APICall
        integration_id = {
            '$connect': 'connect-integration-id',
            '$disconnect': 'disconnect-integration-id',
        }.get(route_key, 'message-integration-id')
        return models.APICall(
            method_name='create_websocket_route',
            params={
                'api_id': Variable('websocket_api_id'),
                'route_key': route_key,
                'integration_id': Variable(integration_id),
            },
        )

    def _plan_websocketapi(self, resource):
        # type: (models.WebsocketAPI) -> Sequence[InstructionMsg]
        configs = self._create_websocket_function_configs(resource)
        routes = resource.routes

        # Which lambda function we use here does not matter. We are only using
        # it to find the account id and the region.
        lambda_arn_var = list(configs.values())[0]['lambda_arn_var']
        shared_plan_preamble = self._arn_parse_instructions(lambda_arn_var) + [
            models.JPSearch('dns_suffix',
                            input_var='parsed_lambda_arn',
                            output_var='dns_suffix'),
        ]  # type: List[InstructionMsg]

        # There's also a set of instructions that are needed
        # at the end of deploying a websocket API that apply to both
        # the update and create case.
        shared_plan_epilogue = [
            models.StoreValue(
                name='websocket_api_url',
                value=StringFormat(
                    'wss://{websocket_api_id}.execute-api.{region_name}'
                    '.{dns_suffix}/%s/' % resource.api_gateway_stage,
                    ['websocket_api_id', 'region_name', 'dns_suffix'],
                ),
            ),
            models.RecordResourceVariable(
                resource_type='websocket_api',
                resource_name=resource.resource_name,
                name='websocket_api_url',
                variable_name='websocket_api_url',
            ),
            models.RecordResourceVariable(
                resource_type='websocket_api',
                resource_name=resource.resource_name,
                name='websocket_api_id',
                variable_name='websocket_api_id',
            ),
        ]  # type: List[InstructionMsg]

        shared_plan_epilogue += [
            models.APICall(
                method_name='add_permission_for_apigateway_v2',
                params={'function_name': function_config['name'],
                        'region_name': Variable('region_name'),
                        'account_id': Variable('account_id'),
                        'api_id': Variable('websocket_api_id')},
            ) for function_config in configs.values()
        ]

        main_plan = []  # type: List[InstructionMsg]
        if not self._remote_state.resource_exists(resource):
            # The resource does not exist, we create it in full here.
            main_plan += [
                (models.APICall(
                    method_name='create_websocket_api',
                    params={'name': resource.name},
                    output_var='websocket_api_id',
                ), "Creating websocket api: %s\n" % resource.name),
                models.StoreValue(
                    name='routes',
                    value=[],
                ),
            ]
            main_plan += self._inject_websocket_integrations(configs)

            for route_key in routes:
                main_plan += [self._create_route_for_key(route_key)]
            main_plan += [
                models.APICall(
                    method_name='deploy_websocket_api',
                    params={
                        'api_id': Variable('websocket_api_id'),
                    },
                    output_var='deployment-id',
                ),
                models.APICall(
                    method_name='create_stage',
                    params={
                        'api_id': Variable('websocket_api_id'),
                        'stage_name': resource.api_gateway_stage,
                        'deployment_id': Variable('deployment-id'),
                    }
                ),
            ]
        else:
            # Already exists. Need to sync up the routes, the easiest way to do
            # this is to delete them and their integrations and re-create them.
            # They will not work if the lambda function changes from under
            # them, and the logic for detecting that and making just the needed
            # changes is complex. There is an integration test to ensure there
            # no dropped messages during a redeployment.
            deployed = self._remote_state.resource_deployed_values(resource)
            main_plan += [
                models.StoreValue(
                    name='websocket_api_id',
                    value=deployed['websocket_api_id']
                ),
                models.APICall(
                    method_name='get_websocket_routes',
                    params={'api_id': Variable('websocket_api_id')},
                    output_var='routes',
                ),
                models.APICall(
                    method_name='delete_websocket_routes',
                    params={
                        'api_id': Variable('websocket_api_id'),
                        'routes': Variable('routes'),
                    },
                ),
                models.APICall(
                    method_name='get_websocket_integrations',
                    params={
                        'api_id': Variable('websocket_api_id'),
                    },
                    output_var='integrations'
                ),
                models.APICall(
                    method_name='delete_websocket_integrations',
                    params={
                        'api_id': Variable('websocket_api_id'),
                        'integrations': Variable('integrations'),
                    }
                )
            ]
            main_plan += self._inject_websocket_integrations(configs)
            for route_key in routes:
                main_plan += [self._create_route_for_key(route_key)]

        ws_plan = shared_plan_preamble + main_plan + shared_plan_epilogue

        if resource.domain_name:
            custom_domain_plan = self._add_custom_domain_plan(
                resource.domain_name, 'REGIONAL',
            )
            ws_plan += custom_domain_plan

        return ws_plan

    def _plan_restapi(self, resource):
        # type: (models.RestAPI) -> Sequence[InstructionMsg]
        function = resource.lambda_function
        function_name = function.function_name
        varname = '%s_lambda_arn' % function.resource_name
        lambda_arn_var = Variable(varname)
        # There's a set of shared instructions that are needed
        # in both the update as well as the initial create case.
        # That's what this shared_plan_premable is for.
        shared_plan_preamble = self._arn_parse_instructions(lambda_arn_var) + [
            models.JPSearch('dns_suffix',
                            input_var='parsed_lambda_arn',
                            output_var='dns_suffix'),
            # The swagger doc uses the 'api_handler_lambda_arn'
            # var name so we need to make sure we populate this variable
            # before importing the rest API.
            models.CopyVariable(from_var=varname,
                                to_var='api_handler_lambda_arn'),
        ]  # type: List[InstructionMsg]
        # There's also a set of instructions that are needed
        # at the end of deploying a rest API that apply to both
        # the update and create case.
        shared_plan_patch_ops = [{
            'op': 'replace',
            'path': '/minimumCompressionSize',
            'value': resource.minimum_compression}
        ]  # type: List[Dict]

        shared_plan_epilogue = [
            models.APICall(
                method_name='update_rest_api',
                params={
                    'rest_api_id': Variable('rest_api_id'),
                    'patch_operations': shared_plan_patch_ops
                }
            ),
            models.APICall(
                method_name='add_permission_for_apigateway',
                params={'function_name': function_name,
                        'region_name': Variable('region_name'),
                        'account_id': Variable('account_id'),
                        'rest_api_id': Variable('rest_api_id')},
            ),
            models.APICall(
                method_name='deploy_rest_api',
                params={'rest_api_id': Variable('rest_api_id'),
                        'xray': resource.xray,
                        'api_gateway_stage': resource.api_gateway_stage},
            ),
            models.StoreValue(
                name='rest_api_url',
                value=StringFormat(
                    'https://{rest_api_id}.execute-api.{region_name}'
                    '.{dns_suffix}/%s/' % resource.api_gateway_stage,
                    ['rest_api_id', 'region_name', 'dns_suffix'],
                ),
            ),
            models.RecordResourceVariable(
                resource_type='rest_api',
                resource_name=resource.resource_name,
                name='rest_api_url',
                variable_name='rest_api_url',
            ),
        ]  # type: List[InstructionMsg]
        for auth in resource.authorizers:
            shared_plan_epilogue.append(
                models.APICall(
                    method_name='add_permission_for_apigateway',
                    params={'function_name': auth.function_name,
                            'region_name': Variable('region_name'),
                            'account_id': Variable('account_id'),
                            'rest_api_id': Variable('rest_api_id')},
                )
            )
        if not self._remote_state.resource_exists(resource):
            plan = shared_plan_preamble + [
                (models.APICall(
                    method_name='import_rest_api',
                    params={'swagger_document': resource.swagger_doc,
                            'endpoint_type': resource.endpoint_type},
                    output_var='rest_api_id',
                ), "Creating Rest API\n"),
                models.RecordResourceVariable(
                    resource_type='rest_api',
                    resource_name=resource.resource_name,
                    name='rest_api_id',
                    variable_name='rest_api_id',
                ),
            ]
        else:
            deployed = self._remote_state.resource_deployed_values(resource)
            shared_plan_epilogue.insert(
                0,
                models.APICall(
                    method_name='get_rest_api',
                    params={'rest_api_id': Variable('rest_api_id')},
                    output_var='rest_api')
            )
            shared_plan_patch_ops.append({
                'op': 'replace',
                'path': StringFormat(
                    '/endpointConfiguration/types/%s' % (
                        '{rest_api[endpointConfiguration][types][0]}'),
                    ['rest_api']),
                'value': resource.endpoint_type}
            )
            plan = shared_plan_preamble + [
                models.StoreValue(
                    name='rest_api_id',
                    value=deployed['rest_api_id']),
                models.RecordResourceVariable(
                    resource_type='rest_api',
                    resource_name=resource.resource_name,
                    name='rest_api_id',
                    variable_name='rest_api_id',
                ),
                (models.APICall(
                    method_name='update_api_from_swagger',
                    params={
                        'rest_api_id': Variable('rest_api_id'),
                        'swagger_document': resource.swagger_doc,
                    },
                ), "Updating rest API\n"),
            ]

        plan.extend(shared_plan_epilogue)

        if resource.domain_name:
            custom_domain_plan = self._add_custom_domain_plan(
                resource.domain_name, resource.endpoint_type
            )
            plan += custom_domain_plan
        return plan

    def _add_custom_domain_plan(self, resource, endpoint_type):
        # type: (models.DomainName, str) -> Sequence[InstructionMsg]
        result = []  # type: List[InstructionMsg]
        custom_domain_plan = self._add_domainname_plan(
            resource, endpoint_type
        )
        result += custom_domain_plan
        api_mapping_plan = self._add_apimapping_plan(
            resource.api_mapping, resource
        )
        result += api_mapping_plan
        return result

    def _get_role_arn(self, resource):
        # type: (models.IAMRole) -> Union[str, Variable]
        if isinstance(resource, models.PreCreatedIAMRole):
            return resource.role_arn
        elif isinstance(resource, models.ManagedIAMRole):
            return Variable('%s_role_arn' % resource.role_name)
        # Make mypy happy.
        raise RuntimeError("Unknown resource type: %s" % resource)

    def _batch_record_resource(self, resource_type, resource_name,
                               mapping):
        # type: (str, str, Dict[str, Any]) -> List[InstructionMsg]
        # This is a helper function for recording multiple values into
        # the same resource dict.  The mapping is the set of variables
        # you want to record.  If the value in a pair is a Variable type,
        # then RecordResourceVariable is used, otherwise, RecordResourceValue
        # is used.
        instructions = []  # type: List[InstructionMsg]
        for key, value in mapping.items():
            instruction = cast(InstructionMsg, None)
            if isinstance(value, Variable):
                instruction = models.RecordResourceVariable(
                    resource_type=resource_type,
                    resource_name=resource_name,
                    name=key,
                    variable_name=value.name
                )
            else:
                instruction = models.RecordResourceValue(
                    resource_type=resource_type,
                    resource_name=resource_name,
                    name=key,
                    value=value
                )
            instructions.append(instruction)
        return instructions


class NoopPlanner(PlanStage):
    def __init__(self):
        # type: () -> None
        pass

    def execute(self, resources):
        # type: (List[models.Model]) -> models.Plan
        return models.Plan(instructions=[], messages={})


class Variable(object):
    def __init__(self, name):
        # type: (str) -> None
        self.name = name

    def __repr__(self):
        # type: () -> str
        return 'Variable("%s")' % self.name

    def __eq__(self, other):
        # type: (Any) -> bool
        return isinstance(other, Variable) and self.name == other.name


class StringFormat(object):
    def __init__(self, template, variables):
        # type: (str, List[str]) -> None
        self.template = template
        self.variables = variables

    def __repr__(self):
        # type: () -> str
        return 'StringFormat("%s")' % self.template

    def __eq__(self, other):
        # type: (Any) -> bool
        return (
            isinstance(other, StringFormat) and
            self.template == other.template and
            self.variables == other.variables
        )


class PlanEncoder(json.JSONEncoder):
    # pylint false positive overriden below
    # https://github.com/PyCQA/pylint/issues/414
    def default(self, o):  # pylint: disable=E0202
        # type: (Any) -> Any
        if isinstance(o, StringFormat):
            return o.template
        return o


class KeyDataVariable(object):
    def __init__(self, name, key):
        # type: (str, str) -> None
        self.name = name
        self.key = key

    def __repr__(self):
        # type: () -> str
        return 'KeyDataVariable("%s", "%s")' % (self.name, self.key)

    def __eq__(self, other):
        # type: (Any) -> bool
        return (
            isinstance(other, KeyDataVariable) and
            self.name == other.name and
            self.key == other.key
        )



from chalice import Chalice

app = Chalice(app_name='{{app_name}}')


@app.route('/')
def index():
    return {'hello': 'world'}


# The view function above will return {"hello": "world"}
# whenever you make an HTTP GET request to '/'.
#
# Here are a few more examples:
#
# @app.route('/hello/{name}')
# def hello_name(name):
#    # '/hello/james' -> {"hello": "james"}
#    return {'hello': name}
#
# @app.route('/users', methods=['POST'])
# def create_user():
#     # This is the JSON body the user sent in their POST request.
#     user_as_json = app.current_request.json_body
#     # We'll echo the json body back to the user in a 'user' key.
#     return {'user': user_as_json}
#
# See the README documentation for more examples.
#



import warnings
try:
    from chalice.cdk.construct import Chalice
except ImportError:
    warnings.warn('Unable to import the Chalice CDK construct due to missing '
                  'dependencies.\nYou can install these by running '
                  "'pip install \"chalice[cdk]\"'")


__all__ = ['Chalice']



import json
import os

from typing import cast
from typing import Dict, List, Tuple, Any, Set, Optional, Text, Union  # noqa
from attr import asdict

from chalice.config import Config  # noqa
from chalice import app
from chalice.constants import LAMBDA_TRUST_POLICY
from chalice.deploy import models
from chalice.utils import UI  # noqa

StrMapAny = Dict[str, Any]


class ChaliceBuildError(Exception):
    pass


class ApplicationGraphBuilder(object):
    def __init__(self):
        # type: () -> None
        self._known_roles = {}  # type: Dict[str, models.IAMRole]
        self._managed_layer = None  # type: Optional[models.LambdaLayer]

    def build(self, config, stage_name):
        # type: (Config, str) -> models.Application
        resources = []  # type: List[models.Model]
        deployment = models.DeploymentPackage(models.Placeholder.BUILD_STAGE)
        for function in config.chalice_app.pure_lambda_functions:
            resource = self._create_lambda_model(
                config=config, deployment=deployment,
                name=function.name, handler_name=function.handler_string,
                stage_name=stage_name)
            resources.append(resource)
        event_resources = self._create_lambda_event_resources(
            config, deployment, stage_name)
        resources.extend(event_resources)
        if config.chalice_app.routes:
            rest_api = self._create_rest_api_model(
                config, deployment, stage_name
            )
            resources.append(rest_api)
        if config.chalice_app.websocket_handlers:
            websocket_api = self._create_websocket_api_model(
                config, deployment, stage_name)
            resources.append(websocket_api)
        return models.Application(stage_name, resources)

    def _create_custom_domain_name(
            self,
            api_type,                # type: models.APIType
            domain_name_data,        # type: StrMapAny
            endpoint_configuration,  # type: str
            api_gateway_stage,       # type: str
    ):
        # type: (...) -> models.DomainName
        url_prefix = domain_name_data.get("url_prefix", '(none)')
        api_mapping_model = self._create_api_mapping_model(
            url_prefix, api_gateway_stage)
        domain_name = self._create_domain_name_model(
            api_type,
            domain_name_data,
            endpoint_configuration,
            api_mapping_model
        )
        return domain_name

    def _create_api_mapping_model(self,
                                  key,         # type: str
                                  stage        # type: str
                                  ):
        # type: (...) -> models.APIMapping
        if key == '/':
            key = '(none)'
        return models.APIMapping(
            resource_name='api_mapping',
            mount_path=key,
            api_gateway_stage=stage
        )

    def _create_lambda_event_resources(self, config, deployment, stage_name):
        # type: (Config, models.DeploymentPackage, str) -> List[models.Model]
        resources = []  # type: List[models.Model]
        for event_source in config.chalice_app.event_sources:
            if isinstance(event_source, app.S3EventConfig):
                resources.append(
                    self._create_bucket_notification(
                        config, deployment, event_source, stage_name
                    )
                )
            elif isinstance(event_source, app.SNSEventConfig):
                resources.append(
                    self._create_sns_subscription(
                        config, deployment, event_source, stage_name,
                    )
                )
            elif isinstance(event_source, app.CloudWatchEventConfig):
                resources.append(
                    self._create_cwe_subscription(
                        config, deployment, event_source, stage_name
                    )
                )
            elif isinstance(event_source, app.ScheduledEventConfig):
                resources.append(
                    self._create_scheduled_model(
                        config, deployment, event_source, stage_name
                    )
                )
            elif isinstance(event_source, app.SQSEventConfig):
                resources.append(
                    self._create_sqs_subscription(
                        config, deployment, event_source, stage_name,
                    )
                )
            elif isinstance(event_source, app.KinesisEventConfig):
                resources.append(
                    self._create_kinesis_subscription(
                        config, deployment, event_source, stage_name,
                    )
                )
            elif isinstance(event_source, app.DynamoDBEventConfig):
                resources.append(
                    self._create_ddb_subscription(
                        config, deployment, event_source, stage_name,
                    )
                )
        return resources

    def _create_rest_api_model(self,
                               config,        # type: Config
                               deployment,    # type: models.DeploymentPackage
                               stage_name,    # type: str
                               ):
        # type: (...) -> models.RestAPI
        # Need to mess with the function name for back-compat.
        lambda_function = self._create_lambda_model(
            config=config, deployment=deployment, name='api_handler',
            handler_name='app.app', stage_name=stage_name
        )
        # For backwards compatibility with the old deployer, the
        # lambda function for the API handler doesn't have the
        # resource_name appended to its complete function_name,
        # it's just <app>-<stage>.
        function_name = '%s-%s' % (config.app_name, config.chalice_stage)
        lambda_function.function_name = function_name
        if config.minimum_compression_size is None:
            minimum_compression = ''
        else:
            minimum_compression = str(config.minimum_compression_size)
        authorizers = []
        for auth in config.chalice_app.builtin_auth_handlers:
            auth_lambda = self._create_lambda_model(
                config=config, deployment=deployment, name=auth.name,
                handler_name=auth.handler_string, stage_name=stage_name,
            )
            authorizers.append(auth_lambda)

        policy = None
        policy_path = config.api_gateway_policy_file
        if (config.api_gateway_endpoint_type == 'PRIVATE' and not policy_path):
            policy = models.IAMPolicy(
                document=self._get_default_private_api_policy(config))
        elif policy_path:
            policy = models.FileBasedIAMPolicy(
                document=models.Placeholder.BUILD_STAGE,
                filename=os.path.join(
                    config.project_dir, '.chalice', policy_path))

        vpce_ids = None
        if config.api_gateway_endpoint_vpce:
            vpce = config.api_gateway_endpoint_vpce
            vpce_ids = [vpce] if isinstance(vpce, str) else vpce

        custom_domain_name = None
        if config.api_gateway_custom_domain:
            custom_domain_name = self._create_custom_domain_name(
                models.APIType.HTTP,
                config.api_gateway_custom_domain,
                config.api_gateway_endpoint_type,
                config.api_gateway_stage,
            )

        return models.RestAPI(
            resource_name='rest_api',
            swagger_doc=models.Placeholder.BUILD_STAGE,
            endpoint_type=config.api_gateway_endpoint_type,
            minimum_compression=minimum_compression,
            api_gateway_stage=config.api_gateway_stage,
            lambda_function=lambda_function,
            authorizers=authorizers,
            policy=policy,
            domain_name=custom_domain_name,
            xray=config.xray_enabled,
            vpce_ids=vpce_ids,
        )

    def _get_default_private_api_policy(self, config):
        # type: (Config) -> StrMapAny
        statements = [{
            "Effect": "Allow",
            "Principal": "*",
            "Action": "execute-api:Invoke",
            "Resource": "arn:*:execute-api:*:*:*",
            "Condition": {
                "StringEquals": {
                    "aws:SourceVpce": config.api_gateway_endpoint_vpce
                }
            }
        }]
        return {"Version": "2012-10-17", "Statement": statements}

    def _create_websocket_api_model(
            self,
            config,      # type: Config
            deployment,  # type: models.DeploymentPackage
            stage_name,  # type: str
    ):
        # type: (...) -> models.WebsocketAPI
        connect_handler = None     # type: Optional[models.LambdaFunction]
        message_handler = None     # type: Optional[models.LambdaFunction]
        disconnect_handler = None  # type: Optional[models.LambdaFunction]

        routes = {h.route_key_handled: h.handler_string for h
                  in config.chalice_app.websocket_handlers.values()}
        if '$connect' in routes:
            connect_handler = self._create_lambda_model(
                config=config, deployment=deployment, name='websocket_connect',
                handler_name=routes['$connect'], stage_name=stage_name)
            routes.pop('$connect')
        if '$disconnect' in routes:
            disconnect_handler = self._create_lambda_model(
                config=config, deployment=deployment,
                name='websocket_disconnect',
                handler_name=routes['$disconnect'], stage_name=stage_name)
            routes.pop('$disconnect')
        if routes:
            # If there are left over routes they are message handlers.
            handler_string = list(routes.values())[0]
            message_handler = self._create_lambda_model(
                config=config, deployment=deployment, name='websocket_message',
                handler_name=handler_string, stage_name=stage_name
            )

        custom_domain_name = None
        if config.websocket_api_custom_domain:
            custom_domain_name = self._create_custom_domain_name(
                models.APIType.WEBSOCKET,
                config.websocket_api_custom_domain,
                config.api_gateway_endpoint_type,
                config.api_gateway_stage,
            )

        return models.WebsocketAPI(
            name='%s-%s-websocket-api' % (config.app_name, stage_name),
            resource_name='websocket_api',
            connect_function=connect_handler,
            message_function=message_handler,
            disconnect_function=disconnect_handler,
            routes=[h.route_key_handled for h
                    in config.chalice_app.websocket_handlers.values()],
            api_gateway_stage=config.api_gateway_stage,
            domain_name=custom_domain_name
        )

    def _create_cwe_subscription(
            self,
            config,        # type: Config
            deployment,    # type: models.DeploymentPackage
            event_source,  # type: app.CloudWatchEventConfig
            stage_name,    # type: str
    ):
        # type: (...) -> models.CloudWatchEvent
        lambda_function = self._create_lambda_model(
            config=config, deployment=deployment, name=event_source.name,
            handler_name=event_source.handler_string, stage_name=stage_name
        )

        resource_name = event_source.name + '-event'
        rule_name = '%s-%s-%s' % (config.app_name, config.chalice_stage,
                                  resource_name)
        cwe = models.CloudWatchEvent(
            resource_name=resource_name,
            rule_name=rule_name,
            event_pattern=json.dumps(event_source.event_pattern),
            lambda_function=lambda_function,
        )
        return cwe

    def _create_scheduled_model(self,
                                config,        # type: Config
                                deployment,    # type: models.DeploymentPackage
                                event_source,  # type: app.ScheduledEventConfig
                                stage_name,    # type: str
                                ):
        # type: (...) -> models.ScheduledEvent
        lambda_function = self._create_lambda_model(
            config=config, deployment=deployment, name=event_source.name,
            handler_name=event_source.handler_string, stage_name=stage_name
        )
        # Resource names must be unique across a chalice app.
        # However, in the original deployer code, the cloudwatch
        # event + lambda function was considered a single resource.
        # Now that they're treated as two separate resources we need
        # a unique name for the event_source that's not the lambda
        # function resource name.  We handle this by just appending
        # '-event' to the name.  Ideally this is handled in app.py
        # but we won't be able to do that until the old deployer
        # is gone.
        resource_name = event_source.name + '-event'
        if isinstance(event_source.schedule_expression,
                      app.ScheduleExpression):
            expression = event_source.schedule_expression.to_string()
        else:
            expression = event_source.schedule_expression
        rule_name = '%s-%s-%s' % (config.app_name, config.chalice_stage,
                                  resource_name)
        scheduled_event = models.ScheduledEvent(
            resource_name=resource_name,
            rule_name=rule_name,
            rule_description=event_source.description,
            schedule_expression=expression,
            lambda_function=lambda_function,
        )
        return scheduled_event

    def _create_domain_name_model(self,
                                  protocol,        # type: models.APIType
                                  data,            # type: StrMapAny
                                  endpoint_type,   # type: str
                                  api_mapping      # type: models.APIMapping
                                  ):
        # type: (...) -> models.DomainName
        default_name = 'api_gateway_custom_domain'
        resource_name_map = {
            'HTTP': default_name,
            'WEBSOCKET': 'websocket_api_custom_domain'
        }  # type: Dict[str, str]

        domain_name = models.DomainName(
            protocol=protocol,
            resource_name=resource_name_map.get(protocol.value, default_name),
            domain_name=data['domain_name'],
            tls_version=models.TLSVersion.create(data.get('tls_version', '')),
            certificate_arn=data['certificate_arn'],
            tags=data.get('tags'),
            api_mapping=api_mapping
        )
        return domain_name

    def _create_lambda_model(self,
                             config,        # type: Config
                             deployment,    # type: models.DeploymentPackage
                             name,          # type: str
                             handler_name,  # type: str
                             stage_name,    # type: str
                             ):
        # type: (...) -> models.LambdaFunction
        new_config = config.scope(
            chalice_stage=config.chalice_stage,
            function_name=name
        )
        role = self._get_role_reference(
            new_config, stage_name, name)
        resource = self._build_lambda_function(
            new_config, name, handler_name,
            deployment, role
        )
        return resource

    def _get_managed_lambda_layer(self, config):
        # type: (Config) -> Optional[models.LambdaLayer]
        if not config.automatic_layer:
            return None
        if self._managed_layer is None:
            self._managed_layer = models.LambdaLayer(
                resource_name='managed-layer',
                layer_name='%s-%s-%s' % (
                    config.app_name, config.chalice_stage, 'managed-layer'),
                runtime=config.lambda_python_version,
                deployment_package=models.DeploymentPackage(
                    models.Placeholder.BUILD_STAGE)
            )
        return self._managed_layer

    def _get_role_reference(self, config, stage_name, function_name):
        # type: (Config, str, str) -> models.IAMRole
        role = self._create_role_reference(config, stage_name, function_name)
        role_identifier = self._get_role_identifier(role)
        if role_identifier in self._known_roles:
            # If we've already create a models.IAMRole with the same
            # identifier, we'll use the existing object instead of
            # creating a new one.
            return self._known_roles[role_identifier]
        self._known_roles[role_identifier] = role
        return role

    def _get_role_identifier(self, role):
        # type: (models.IAMRole) -> str
        if isinstance(role, models.PreCreatedIAMRole):
            return role.role_arn
        # We know that if it's not a PreCreatedIAMRole, it's
        # a managed role, so we're using cast() to make mypy happy.
        role = cast(models.ManagedIAMRole, role)
        return role.resource_name

    def _create_role_reference(self, config, stage_name, function_name):
        # type: (Config, str, str) -> models.IAMRole
        # First option, the user doesn't want us to manage
        # the role at all.
        if not config.manage_iam_role:
            # We've already validated the iam_role_arn is provided
            # if manage_iam_role is set to False.
            return models.PreCreatedIAMRole(
                role_arn=config.iam_role_arn,
            )
        policy = models.IAMPolicy(document=models.Placeholder.BUILD_STAGE)
        if not config.autogen_policy:
            resource_name = '%s_role' % function_name
            role_name = '%s-%s-%s' % (config.app_name, stage_name,
                                      function_name)
            if config.iam_policy_file is not None:
                filename = os.path.join(config.project_dir,
                                        '.chalice',
                                        config.iam_policy_file)
            else:
                filename = os.path.join(config.project_dir,
                                        '.chalice',
                                        'policy-%s.json' % stage_name)
            policy = models.FileBasedIAMPolicy(
                filename=filename, document=models.Placeholder.BUILD_STAGE)
        else:
            resource_name = 'default-role'
            role_name = '%s-%s' % (config.app_name, stage_name)
            policy = models.AutoGenIAMPolicy(
                document=models.Placeholder.BUILD_STAGE,
                traits=set([]),
            )
        return models.ManagedIAMRole(
            resource_name=resource_name,
            role_name=role_name,
            trust_policy=LAMBDA_TRUST_POLICY,
            policy=policy,
        )

    def _get_vpc_params(self, function_name, config):
        # type: (str, Config) -> Tuple[List[str], List[str]]
        security_group_ids = config.security_group_ids
        subnet_ids = config.subnet_ids
        if security_group_ids and subnet_ids:
            return security_group_ids, subnet_ids
        elif not security_group_ids and not subnet_ids:
            return [], []
        else:
            raise ChaliceBuildError(
                "Invalid VPC params for function '%s', in order to configure "
                "VPC for a Lambda function, you must provide the subnet_ids "
                "as well as the security_group_ids, got subnet_ids: %s, "
                "security_group_ids: %s" % (function_name,
                                            subnet_ids,
                                            security_group_ids)
            )

    def _get_lambda_layers(self, config):
        # type: (Config) -> List[str]
        layers = config.layers
        return layers if layers else []

    def _build_lambda_function(self,
                               config,        # type: Config
                               name,          # type: str
                               handler_name,  # type: str
                               deployment,    # type: models.DeploymentPackage
                               role,          # type: models.IAMRole
                               ):
        # type: (...) -> models.LambdaFunction
        function_name = '%s-%s-%s' % (
            config.app_name, config.chalice_stage, name)
        security_group_ids, subnet_ids = self._get_vpc_params(name, config)
        lambda_layers = self._get_lambda_layers(config)
        function = models.LambdaFunction(
            resource_name=name,
            function_name=function_name,
            environment_variables=config.environment_variables,
            runtime=config.lambda_python_version,
            handler=handler_name,
            tags=config.tags,
            timeout=config.lambda_timeout,
            memory_size=config.lambda_memory_size,
            deployment_package=deployment,
            role=role,
            security_group_ids=security_group_ids,
            subnet_ids=subnet_ids,
            reserved_concurrency=config.reserved_concurrency,
            layers=lambda_layers,
            managed_layer=self._get_managed_lambda_layer(config),
            xray=config.xray_enabled,
        )
        self._inject_role_traits(function, role)
        return function

    def _inject_role_traits(self, function, role):
        # type: (models.LambdaFunction, models.IAMRole) -> None
        if not isinstance(role, models.ManagedIAMRole):
            return
        policy = role.policy
        if not isinstance(policy, models.AutoGenIAMPolicy):
            return
        if function.security_group_ids and function.subnet_ids:
            policy.traits.add(models.RoleTraits.VPC_NEEDED)

    def _create_bucket_notification(
        self,
        config,      # type: Config
        deployment,  # type: models.DeploymentPackage
        s3_event,    # type: app.S3EventConfig
        stage_name,  # type: str
    ):
        # type: (...) -> models.S3BucketNotification
        lambda_function = self._create_lambda_model(
            config=config, deployment=deployment, name=s3_event.name,
            handler_name=s3_event.handler_string, stage_name=stage_name
        )
        resource_name = s3_event.name + '-s3event'
        s3_bucket = models.S3BucketNotification(
            resource_name=resource_name,
            bucket=s3_event.bucket,
            prefix=s3_event.prefix,
            suffix=s3_event.suffix,
            events=s3_event.events,
            lambda_function=lambda_function,
        )
        return s3_bucket

    def _create_sns_subscription(
        self,
        config,      # type: Config
        deployment,  # type: models.DeploymentPackage
        sns_config,  # type: app.SNSEventConfig
        stage_name,  # type: str
    ):
        # type: (...) -> models.SNSLambdaSubscription
        lambda_function = self._create_lambda_model(
            config=config, deployment=deployment, name=sns_config.name,
            handler_name=sns_config.handler_string, stage_name=stage_name
        )
        resource_name = sns_config.name + '-sns-subscription'
        sns_subscription = models.SNSLambdaSubscription(
            resource_name=resource_name,
            topic=sns_config.topic,
            lambda_function=lambda_function,
        )
        return sns_subscription

    def _create_sqs_subscription(
        self,
        config,      # type: Config
        deployment,  # type: models.DeploymentPackage
        sqs_config,  # type: app.SQSEventConfig
        stage_name,  # type: str
    ):
        # type: (...) -> models.SQSEventSource
        lambda_function = self._create_lambda_model(
            config=config, deployment=deployment, name=sqs_config.name,
            handler_name=sqs_config.handler_string, stage_name=stage_name
        )
        resource_name = sqs_config.name + '-sqs-event-source'
        queue = ''  # type: Union[str, models.QueueARN]
        if sqs_config.queue_arn is not None:
            queue = models.QueueARN(arn=sqs_config.queue_arn)
        elif sqs_config.queue is not None:
            queue = sqs_config.queue
        batch_window = sqs_config.maximum_batching_window_in_seconds
        sqs_event_source = models.SQSEventSource(
            resource_name=resource_name,
            queue=queue,
            batch_size=sqs_config.batch_size,
            lambda_function=lambda_function,
            maximum_batching_window_in_seconds=batch_window,
        )
        return sqs_event_source

    def _create_kinesis_subscription(
        self,
        config,          # type: Config
        deployment,      # type: models.DeploymentPackage
        kinesis_config,  # type: app.KinesisEventConfig
        stage_name,      # type: str
    ):
        # type: (...) -> models.KinesisEventSource
        lambda_function = self._create_lambda_model(
            config=config, deployment=deployment, name=kinesis_config.name,
            handler_name=kinesis_config.handler_string, stage_name=stage_name
        )
        resource_name = kinesis_config.name + '-kinesis-event-source'
        batch_window = kinesis_config.maximum_batching_window_in_seconds
        kinesis_event_source = models.KinesisEventSource(
            resource_name=resource_name,
            stream=kinesis_config.stream,
            batch_size=kinesis_config.batch_size,
            maximum_batching_window_in_seconds=batch_window,
            starting_position=kinesis_config.starting_position,
            lambda_function=lambda_function,
        )
        return kinesis_event_source

    def _create_ddb_subscription(
        self,
        config,          # type: Config
        deployment,      # type: models.DeploymentPackage
        ddb_config,      # type: app.DynamoDBEventConfig
        stage_name,      # type: str
    ):
        # type: (...) -> models.DynamoDBEventSource
        lambda_function = self._create_lambda_model(
            config=config, deployment=deployment, name=ddb_config.name,
            handler_name=ddb_config.handler_string, stage_name=stage_name
        )
        resource_name = ddb_config.name + '-dynamodb-event-source'
        batch_window = ddb_config.maximum_batching_window_in_seconds
        ddb_event_source = models.DynamoDBEventSource(
            resource_name=resource_name,
            stream_arn=ddb_config.stream_arn,
            batch_size=ddb_config.batch_size,
            maximum_batching_window_in_seconds=batch_window,
            starting_position=ddb_config.starting_position,
            lambda_function=lambda_function,
        )
        return ddb_event_source


class DependencyBuilder(object):
    def __init__(self):
        # type: () -> None
        pass

    def build_dependencies(self, graph):
        # type: (models.Model) -> List[models.Model]
        seen = set()  # type: Set[int]
        ordered = []  # type: List[models.Model]
        for resource in graph.dependencies():
            self._traverse(resource, ordered, seen)
        return ordered

    def _traverse(self, resource, ordered, seen):
        # type: (models.Model, List[models.Model], Set[int]) -> None
        for dep in resource.dependencies():
            if id(dep) not in seen:
                seen.add(id(dep))
                self._traverse(dep, ordered, seen)
        # If recreating this list is a perf issue later on,
        # we can create yet-another set of ids that gets updated
        # when we add a resource to the ordered list.
        if id(resource) not in [id(r) for r in ordered]:
            ordered.append(resource)


class GraphPrettyPrint(object):

    _NEW_SECTION = '\u251c\u2500\u2500'
    _LINE_VERTICAL = '\u2502'

    def __init__(self, ui):
        # type: (UI) -> None
        self._ui = ui

    def display_graph(self, graph):
        # type: (models.Model) -> None
        self._ui.write("Application\n")
        for model in graph.dependencies():
            self._traverse(model, level=0)

    def _traverse(self, graph, level):
        # type: (models.Model, int) -> None
        prefix = ('%s   ' % self._LINE_VERTICAL) * level
        spaces = prefix + self._NEW_SECTION + ' '
        model_text = self._get_model_text(graph, spaces, level)
        current_line = cast(str, '%s%s\n' % (spaces, model_text))
        self._ui.write(current_line)
        for model in graph.dependencies():
            self._traverse(model, level + 1)

    def _get_model_text(self, model, spaces, level):
        # type: (models.Model, Text, int) -> Text
        name = model.__class__.__name__
        filtered = self._get_filtered_params(model)
        if not filtered:
            return '%s()' % name
        total_len_prefix = len(spaces) + len(name) + 1
        prefix = ('%s   ' % self._LINE_VERTICAL) * (level + 2)
        full = '%s%s' % (prefix, ' ' * (total_len_prefix - len(prefix)))
        param_items = list(filtered.items())
        first = param_items[0]
        remaining = param_items[1:]
        lines = ['%s(%s=%s,' % (name, first[0], first[1])]
        self._add_remaining_lines(lines, remaining, full)
        return '\n'.join(lines) + ')'

    def _add_remaining_lines(self, lines, remaining, full):
        # type: (List[str], List[Tuple[str, Any]], Text) -> None
        for key, value in remaining:
            if isinstance(value, (list, dict)):
                value = key.upper()
            current = cast(str, '%s%s=%s,' % (full, key, value))
            lines.append(current)

    def _get_filtered_params(self, model):
        # type: (models.Model) -> StrMapAny
        dependencies = model.dependencies()
        filtered = asdict(
            model, filter=lambda _, v: v not in dependencies and v)
        return filtered



"""Dev server used for running a chalice app locally.

This is intended only for local development purposes.

"""
from __future__ import print_function
import re
import threading
import time
import uuid
import base64
import functools
import warnings
from collections import namedtuple
import json

from six.moves.BaseHTTPServer import HTTPServer
from six.moves.BaseHTTPServer import BaseHTTPRequestHandler
from six.moves.socketserver import ThreadingMixIn
from typing import (
    List,
    Any,
    Dict,
    Tuple,
    Callable,
    Optional,
    Union,
)  # noqa

from chalice.app import Chalice  # noqa
from chalice.app import CORSConfig  # noqa
from chalice.app import ChaliceAuthorizer  # noqa
from chalice.app import CognitoUserPoolAuthorizer  # noqa
from chalice.app import RouteEntry  # noqa
from chalice.app import Request  # noqa
from chalice.app import AuthResponse  # noqa
from chalice.app import BuiltinAuthConfig  # noqa
from chalice.config import Config  # noqa

from chalice.compat import urlparse, parse_qs


MatchResult = namedtuple('MatchResult', ['route', 'captured', 'query_params'])
EventType = Dict[str, Any]
ContextType = Dict[str, Any]
HeaderType = Dict[str, Any]
ResponseType = Dict[str, Any]
HandlerCls = Callable[..., 'ChaliceRequestHandler']
ServerCls = Callable[..., 'HTTPServer']


class Clock(object):
    def time(self):
        # type: () -> float
        return time.time()


def create_local_server(app_obj, config, host, port):
    # type: (Chalice, Config, str, int) -> LocalDevServer
    CustomLocalChalice.__bases__ = (LocalChalice, app_obj.__class__)
    app_obj.__class__ = CustomLocalChalice
    return LocalDevServer(app_obj, config, host, port)


class LocalARNBuilder(object):
    ARN_FORMAT = ('arn:aws:execute-api:{region}:{account_id}'
                  ':{api_id}/{stage}/{method}/{resource_path}')
    LOCAL_REGION = 'mars-west-1'
    LOCAL_ACCOUNT_ID = '123456789012'
    LOCAL_API_ID = 'ymy8tbxw7b'
    LOCAL_STAGE = 'api'

    def build_arn(self, method, path):
        # type: (str, str) -> str
        # In API Gateway the method and URI are separated by a / so typically
        # the uri portion omits the leading /. In the case where the entire
        # url is just '/' API Gateway adds a / to the end so that the arn end
        # with a '//'.
        if path != '/':
            path = path[1:]
        path = path.split('?')[0]
        return self.ARN_FORMAT.format(
            region=self.LOCAL_REGION,
            account_id=self.LOCAL_ACCOUNT_ID,
            api_id=self.LOCAL_API_ID,
            stage=self.LOCAL_STAGE,
            method=method,
            resource_path=path
        )


class ARNMatcher(object):
    def __init__(self, target_arn):
        # type: (str) -> None
        self._arn = target_arn

    def _resource_match(self, resource):
        # type: (str) -> bool
        # Arn matching supports two special case characetrs that are not
        # escapable. * represents a glob which translates to a non-greedy
        # match of any number of characters. ? which is any single character.
        # These are easy to translate to a regex using .*? and . respectivly.
        escaped_resource = re.escape(resource)
        resource_regex = escaped_resource.replace(r'\?', '.').replace(
            r'\*', '.*?')
        resource_regex = '^%s$' % resource_regex
        return re.match(resource_regex, self._arn) is not None

    def does_any_resource_match(self, resources):
        # type: (List[str]) -> bool
        for resource in resources:
            if self._resource_match(resource):
                return True
        return False


class RouteMatcher(object):
    def __init__(self, route_urls):
        # type: (List[str]) -> None
        # Sorting the route_urls ensures we always check
        # the concrete routes for a prefix before the
        # variable/capture parts of the route, e.g
        # '/foo/bar' before '/foo/{capture}'
        self.route_urls = sorted(route_urls)

    def match_route(self, url):
        # type: (str) -> MatchResult
        """Match the url against known routes.

        This method takes a concrete route "/foo/bar", and
        matches it against a set of routes.  These routes can
        use param substitution corresponding to API gateway patterns.
        For example::

            match_route('/foo/bar') -> '/foo/{name}'

        """
        # Otherwise we need to check for param substitution
        parsed_url = urlparse(url)
        query_params = parse_qs(parsed_url.query, keep_blank_values=True)
        path = parsed_url.path
        # API Gateway removes the trailing slash if the route is not the root
        # path. We do the same here so our route matching works the same way.
        if path != '/' and path.endswith('/'):
            path = path[:-1]
        parts = path.split('/')
        captured = {}
        for route_url in self.route_urls:
            url_parts = route_url.split('/')
            if len(parts) == len(url_parts):
                for i, j in zip(parts, url_parts):
                    if j.startswith('{') and j.endswith('}'):
                        captured[j[1:-1]] = i
                        continue
                    if i != j:
                        break
                else:
                    return MatchResult(route_url, captured, query_params)
        raise ValueError("No matching route found for: %s" % url)


class LambdaEventConverter(object):

    LOCAL_SOURCE_IP = '127.0.0.1'

    """Convert an HTTP request to an event dict used by lambda."""
    def __init__(self, route_matcher, binary_types=None):
        # type: (RouteMatcher, List[str]) -> None
        self._route_matcher = route_matcher
        if binary_types is None:
            binary_types = []
        self._binary_types = binary_types

    def _is_binary(self, headers):
        # type: (Dict[str,Any]) -> bool
        return headers.get('content-type', '') in self._binary_types

    def create_lambda_event(self, method, path, headers, body=None):
        # type: (str, str, Dict[str, str], bytes) -> EventType
        view_route = self._route_matcher.match_route(path)
        event = {
            'requestContext': {
                'httpMethod': method,
                'resourcePath': view_route.route,
                'identity': {
                    'sourceIp': self.LOCAL_SOURCE_IP
                },
                'path': path.split('?')[0],
            },
            'headers': {k.lower(): v for k, v in headers.items()},
            'pathParameters': view_route.captured,
            'stageVariables': {},
        }
        if view_route.query_params:
            event['multiValueQueryStringParameters'] = view_route.query_params
        else:
            # If no query parameters are provided, API gateway maps
            # this to None so we're doing this for parity.
            event['multiValueQueryStringParameters'] = None
        if self._is_binary(headers) and body is not None:
            event['body'] = base64.b64encode(body).decode('ascii')
            event['isBase64Encoded'] = True
        else:
            event['body'] = body
        return event


class LocalGatewayException(Exception):
    CODE = 0

    def __init__(self, headers, body=None):
        # type: (HeaderType, Optional[bytes]) -> None
        self.headers = headers
        self.body = body


class InvalidAuthorizerError(LocalGatewayException):
    CODE = 500


class ForbiddenError(LocalGatewayException):
    CODE = 403


class NotAuthorizedError(LocalGatewayException):
    CODE = 401


class LambdaContext(object):
    def __init__(self, function_name, memory_size,
                 max_runtime_ms=3000, time_source=None):
        # type: (str, int, int, Optional[Clock]) -> None
        if time_source is None:
            time_source = Clock()
        self._time_source = time_source
        self._start_time = self._current_time_millis()
        self._max_runtime = max_runtime_ms

        # Below are properties that are found on the real LambdaContext passed
        # by lambda and their associated documentation.

        # Name of the Lambda function that is executing.
        self.function_name = function_name

        # The Lambda function version that is executing. If an alias is used
        # to invoke the function, then function_version will be the version
        # the alias points to.
        # Chalice local obviously does not support versioning so it will always
        # be set to $LATEST.
        self.function_version = '$LATEST'

        # The ARN used to invoke this function. It can be function ARN or
        # alias ARN. An unqualified ARN executes the $LATEST version and
        # aliases execute the function version it is pointing to.
        self.invoked_function_arn = ''

        # Memory limit, in MB, you configured for the Lambda function. You set
        # the memory limit at the time you create a Lambda function and you
        # can change it later.
        self.memory_limit_in_mb = memory_size

        # AWS request ID associated with the request. This is the ID returned
        # to the client that called the invoke method.
        self.aws_request_id = str(uuid.uuid4())

        # The name of the CloudWatch log group where you can find logs written
        # by your Lambda function.
        self.log_group_name = ''

        # The name of the CloudWatch log stream where you can find logs
        # written by your Lambda function. The log stream may or may not
        # change for each invocation of the Lambda function.
        #
        # The value is null if your Lambda function is unable to create a log
        # stream, which can happen if the execution role that grants necessary
        # permissions to the Lambda function does not include permissions for
        # the CloudWatch Logs actions.
        self.log_stream_name = ''

        # The last two attributes have the following comment in the
        # documentation:
        # Information about the client application and device when invoked
        # through the AWS Mobile SDK, it can be null.
        # Chalice local doens't need to set these since they are specifically
        # for the mobile SDK.
        self.identity = None
        self.client_context = None

    def _current_time_millis(self):
        # type: () -> float
        return self._time_source.time() * 1000

    def get_remaining_time_in_millis(self):
        # type: () -> float
        runtime = self._current_time_millis() - self._start_time
        return self._max_runtime - runtime


LocalAuthPair = Tuple[EventType, LambdaContext]


class LocalGatewayAuthorizer(object):
    """A class for running user defined authorizers in local mode."""
    def __init__(self, app_object):
        # type: (Chalice) -> None
        self._app_object = app_object
        self._arn_builder = LocalARNBuilder()

    def authorize(self, raw_path, lambda_event, lambda_context):
        # type: (str, EventType, LambdaContext) -> LocalAuthPair
        method = lambda_event['requestContext']['httpMethod']
        route_entry = self._route_for_event(lambda_event)
        if not route_entry:
            return lambda_event, lambda_context
        authorizer = route_entry.authorizer
        if not authorizer:
            return lambda_event, lambda_context
        # If authorizer is Cognito then try to parse the JWT and simulate an
        # APIGateway validated request
        if isinstance(authorizer, CognitoUserPoolAuthorizer):
            if "headers" in lambda_event\
                    and "authorization" in lambda_event["headers"]:
                token = lambda_event["headers"]["authorization"]
                claims = self._decode_jwt_payload(token)

                try:
                    cognito_username = claims["cognito:username"]
                except KeyError:
                    # If a key error is raised when trying to get the cognito
                    # username then it is a machine-to-machine communication.
                    # This kind of cognito authorization flow is not
                    # supported in local mode. We can ignore it here to allow
                    # users to test their code local with a different cognito
                    # authorization flow.
                    warnings.warn(
                        '%s for machine-to-machine communicaiton is not '
                        'supported in local mode. All requests made against '
                        'a route will be authorized to allow local testing.'
                        % authorizer.__class__.__name__
                    )
                    return lambda_event, lambda_context

                auth_result = {"context": {"claims": claims},
                               "principalId": cognito_username}
                lambda_event = self._update_lambda_event(lambda_event,
                                                         auth_result)
        if not isinstance(authorizer, ChaliceAuthorizer):
            # Currently the only supported local authorizer is the
            # BuiltinAuthConfig type. Anything else we will err on the side of
            # allowing local testing by simply admiting the request. Otherwise
            # there is no way for users to test their code in local mode.
            warnings.warn(
                '%s is not a supported in local mode. All requests made '
                'against a route will be authorized to allow local testing.'
                % authorizer.__class__.__name__
            )
            return lambda_event, lambda_context
        arn = self._arn_builder.build_arn(method, raw_path)
        auth_event = self._prepare_authorizer_event(arn, lambda_event,
                                                    lambda_context)
        auth_result = authorizer(auth_event, lambda_context)
        if auth_result is None:
            raise InvalidAuthorizerError(
                {'x-amzn-RequestId': lambda_context.aws_request_id,
                 'x-amzn-ErrorType': 'AuthorizerConfigurationException'},
                b'{"message":null}'
            )
        authed = self._check_can_invoke_view_function(arn, auth_result)
        if authed:
            lambda_event = self._update_lambda_event(lambda_event, auth_result)
        else:
            raise ForbiddenError(
                {'x-amzn-RequestId': lambda_context.aws_request_id,
                 'x-amzn-ErrorType': 'AccessDeniedException'},
                (b'{"Message": '
                 b'"User is not authorized to access this resource"}'))
        return lambda_event, lambda_context

    def _check_can_invoke_view_function(self, arn, auth_result):
        # type: (str, ResponseType) -> bool
        policy = auth_result.get('policyDocument', {})
        statements = policy.get('Statement', [])
        allow_resource_statements = []
        for statement in statements:
            if statement.get('Effect') == 'Allow' and \
                    (statement.get('Action') == 'execute-api:Invoke' or
                     'execute-api:Invoke' in statement.get('Action')):
                for resource in statement.get('Resource'):
                    allow_resource_statements.append(resource)

        arn_matcher = ARNMatcher(arn)
        return arn_matcher.does_any_resource_match(allow_resource_statements)

    def _route_for_event(self, lambda_event):
        # type: (EventType) -> Optional[RouteEntry]
        # Authorizer had to be made into an Any type since mypy couldn't
        # detect that app.ChaliceAuthorizer was callable.
        resource_path = lambda_event.get(
            'requestContext', {}).get('resourcePath')
        http_method = lambda_event['requestContext']['httpMethod']
        try:
            route_entry = self._app_object.routes[resource_path][http_method]
        except KeyError:
            # If a key error is raised when trying to get the route entry
            # then this route does not support this method. A method error
            # will be raised by the chalice handler method. We can ignore it
            # here by returning no authorizer to avoid duplicating the logic.
            return None
        return route_entry

    def _update_lambda_event(self, lambda_event, auth_result):
        # type: (EventType, ResponseType) -> EventType
        auth_context = auth_result['context']
        auth_context.update({
            'principalId': auth_result['principalId']
        })
        lambda_event['requestContext']['authorizer'] = auth_context
        return lambda_event

    def _prepare_authorizer_event(self, arn, lambda_event, lambda_context):
        # type: (str, EventType, LambdaContext) -> EventType
        """Translate event for an authorizer input."""
        authorizer_event = lambda_event.copy()
        authorizer_event['type'] = 'TOKEN'
        try:
            authorizer_event['authorizationToken'] = authorizer_event.get(
                'headers', {})['authorization']
        except KeyError:
            raise NotAuthorizedError(
                {'x-amzn-RequestId': lambda_context.aws_request_id,
                 'x-amzn-ErrorType': 'UnauthorizedException'},
                b'{"message":"Unauthorized"}')
        authorizer_event['methodArn'] = arn
        return authorizer_event

    def _decode_jwt_payload(self, jwt):
        # type: (str) -> Dict
        payload_segment = jwt.split(".", 2)[1]
        payload = base64.urlsafe_b64decode(self._base64_pad(payload_segment))
        return json.loads(payload)

    def _base64_pad(self, value):
        # type: (str) -> str
        rem = len(value) % 4
        if rem > 0:
            value += "=" * (4 - rem)
        return value


class LocalGateway(object):
    """A class for faking the behavior of API Gateway."""
    def __init__(self, app_object, config):
        # type: (Chalice, Config) -> None
        self._app_object = app_object
        self._config = config
        self.event_converter = LambdaEventConverter(
            RouteMatcher(list(app_object.routes)),
            self._app_object.api.binary_types
        )
        self._authorizer = LocalGatewayAuthorizer(app_object)

    def _generate_lambda_context(self):
        # type: () -> LambdaContext
        if self._config.lambda_timeout is None:
            timeout = None
        else:
            timeout = self._config.lambda_timeout * 1000
        return LambdaContext(
            function_name=self._config.function_name,
            memory_size=self._config.lambda_memory_size,
            max_runtime_ms=timeout
        )

    def _generate_lambda_event(self, method, path, headers, body):
        # type: (str, str, HeaderType, Optional[bytes]) -> EventType
        lambda_event = self.event_converter.create_lambda_event(
            method=method, path=path, headers=headers,
            body=body,
        )
        return lambda_event

    def _has_user_defined_options_method(self, lambda_event):
        # type: (EventType) -> bool
        route_key = lambda_event['requestContext']['resourcePath']
        return 'OPTIONS' in self._app_object.routes[route_key]

    def handle_request(self, method, path, headers, body):
        # type: (str, str, HeaderType, Optional[bytes]) -> ResponseType
        lambda_context = self._generate_lambda_context()
        try:
            lambda_event = self._generate_lambda_event(
                method, path, headers, body)
        except ValueError:
            # API Gateway will return a different error on route not found
            # depending on whether or not we have an authorization token in our
            # request. Since we do not do that check until we actually find
            # the authorizer that we will call we do not have that information
            # available at this point. Instead we just check to see if that
            # header is present and change our response if it is. This will
            # need to be refactored later if we decide to more closely mirror
            # how API Gateway does their auth and routing.
            error_headers = {'x-amzn-RequestId': lambda_context.aws_request_id,
                             'x-amzn-ErrorType': 'UnauthorizedException'}
            auth_header = headers.get('authorization')
            if auth_header is None:
                auth_header = headers.get('Authorization')
            if auth_header is not None:
                raise ForbiddenError(
                    error_headers,
                    (b'{"message": "Authorization header requires '
                     b'\'Credential\''
                     b' parameter. Authorization header requires \'Signature\''
                     b' parameter. Authorization header requires '
                     b'\'SignedHeaders\' parameter. Authorization header '
                     b'requires existence of either a \'X-Amz-Date\' or a'
                     b' \'Date\' header. Authorization=%s"}'
                     % auth_header.encode('ascii')))
            raise ForbiddenError(
                error_headers,
                b'{"message": "Missing Authentication Token"}')

        # This can either be because the user's provided an OPTIONS method
        # *or* this is a preflight request, which chalice automatically
        # responds to without invoking a user defined route.
        if method == 'OPTIONS' and \
           not self._has_user_defined_options_method(lambda_event):
            # No options route was defined for this path. API Gateway should
            # automatically generate our CORS headers.
            options_headers = self._autogen_options_headers(lambda_event)
            return {
                'statusCode': 200,
                'headers': options_headers,
                'multiValueHeaders': {},
                'body': None
            }
        # The authorizer call will be a noop if there is no authorizer method
        # defined for route. Otherwise it will raise a ForbiddenError
        # which will be caught by the handler that called this and a 403 or
        # 401 will be sent back over the wire.
        lambda_event, lambda_context = self._authorizer.authorize(
            path, lambda_event, lambda_context)
        response = self._app_object(lambda_event, lambda_context)
        return response

    def _autogen_options_headers(self, lambda_event):
        # type:(EventType) -> HeaderType
        route_key = lambda_event['requestContext']['resourcePath']
        route_dict = self._app_object.routes[route_key]
        route_methods = [method for method in route_dict.keys()
                         if route_dict[method].cors is not None]

        # If there are no views with CORS enabled
        # then OPTIONS is the only allowed method.
        if not route_methods:
            return {'Access-Control-Allow-Methods': 'OPTIONS'}

        # Chalice ensures that routes with multiple views have the same
        # CORS configuration, so if any view has a CORS Config we can use
        # that config since they will all be the same.
        cors_config = route_dict[route_methods[0]].cors
        cors_headers = cors_config.get_access_control_headers()

        # We need to add OPTIONS since it is not a part of the CORSConfig
        # object. APIGateway handles this entirely based on the API definition.
        # So our local version needs to add this manually to our set of allowed
        # headers.
        route_methods.append('OPTIONS')

        # The Access-Control-Allow-Methods header is not added by the
        # CORSConfig object it is added to the API Gateway route during
        # deployment, so we need to manually add those headers here.
        cors_headers.update({
            'Access-Control-Allow-Methods': '%s' % ','.join(route_methods)
        })
        return cors_headers


class ChaliceRequestHandler(BaseHTTPRequestHandler):
    """A class for mapping raw HTTP events to and from LocalGateway."""
    protocol_version = 'HTTP/1.1'

    def __init__(self, request, client_address, server, app_object, config):
        # type: (bytes, Tuple[str, int], HTTPServer, Chalice, Config) -> None
        self.local_gateway = LocalGateway(app_object, config)
        BaseHTTPRequestHandler.__init__(
            self, request, client_address, server)  # type: ignore

    def _parse_payload(self):
        # type: () -> Tuple[HeaderType, Optional[bytes]]
        body = None
        content_length = int(self.headers.get('content-length', '0'))
        if content_length > 0:
            body = self.rfile.read(content_length)
        converted_headers = dict(self.headers)
        return converted_headers, body

    def _generic_handle(self):
        # type: () -> None
        headers, body = self._parse_payload()
        try:
            response = self.local_gateway.handle_request(
                method=self.command,
                path=self.path,
                headers=headers,
                body=body
            )
            status_code = response['statusCode']
            headers = response['headers'].copy()
            headers.update(response['multiValueHeaders'])
            response = self._handle_binary(response)
            body = response['body']
            self._send_http_response(status_code, headers, body)
        except LocalGatewayException as e:
            self._send_error_response(e)

    def _handle_binary(self, response):
        # type: (Dict[str,Any]) -> Dict[str,Any]
        if response.get('isBase64Encoded'):
            body = base64.b64decode(response['body'])
            response['body'] = body
        return response

    def _send_error_response(self, error):
        # type: (LocalGatewayException) -> None
        code = error.CODE
        headers = error.headers
        body = error.body
        self._send_http_response(code, headers, body)

    def _send_http_response(self, code, headers, body):
        # type: (int, HeaderType, Optional[Union[str,bytes]]) -> None
        if body is None:
            self._send_http_response_no_body(code, headers)
        else:
            self._send_http_response_with_body(code, headers, body)

    def _send_http_response_with_body(self, code, headers, body):
        # type: (int, HeaderType, Union[str,bytes]) -> None
        self.send_response(code)
        if not isinstance(body, bytes):
            body = body.encode('utf-8')
        self.send_header('Content-Length', str(len(body)))
        content_type = headers.pop(
            'Content-Type', 'application/json')
        self.send_header('Content-Type', content_type)
        self._send_headers(headers)
        self.wfile.write(body)

    do_GET = do_PUT = do_POST = do_HEAD = do_DELETE = do_PATCH = do_OPTIONS = \
        _generic_handle

    def _send_http_response_no_body(self, code, headers):
        # type: (int, HeaderType) -> None
        headers['Content-Length'] = '0'
        self.send_response(code)
        self._send_headers(headers)

    def _send_headers(self, headers):
        # type: (HeaderType) -> None
        for header_name, header_value in headers.items():
            if isinstance(header_value, list):
                for value in header_value:
                    self.send_header(header_name, value)
            else:
                self.send_header(header_name, header_value)
        self.end_headers()


class ThreadedHTTPServer(ThreadingMixIn, HTTPServer):
    """Threading mixin to better support browsers.

    When a browser sends a GET request to Chalice it keeps the connection open
    for reuse. In the single threaded model this causes Chalice local to become
    unresponsive to all clients other than that browser socket. Even sending a
    header requesting that the client close the connection is not good enough,
    the browswer will simply open another one and sit on it.
    """

    daemon_threads = True


class LocalDevServer(object):
    def __init__(self, app_object, config, host, port,
                 handler_cls=ChaliceRequestHandler,
                 server_cls=ThreadedHTTPServer):
        # type: (Chalice, Config, str, int, HandlerCls, ServerCls) -> None
        self.app_object = app_object
        self.host = host
        self.port = port
        self._wrapped_handler = functools.partial(
            handler_cls, app_object=app_object, config=config)
        self.server = server_cls((host, port), self._wrapped_handler)

    def handle_single_request(self):
        # type: () -> None
        self.server.handle_request()

    def serve_forever(self):
        # type: () -> None
        print("Serving on http://%s:%s" % (self.host, self.port))
        self.server.serve_forever()

    def shutdown(self):
        # type: () -> None
        # This must be called from another thread of else it
        # will deadlock.
        self.server.shutdown()


class HTTPServerThread(threading.Thread):
    """Thread that manages starting/stopping local HTTP server.

    This is a small wrapper around a normal threading.Thread except
    that it adds shutdown capability of the HTTP server, which is
    not part of the normal threading.Thread interface.

    """
    def __init__(self, server_factory):
        # type: (Callable[[], LocalDevServer]) -> None
        threading.Thread.__init__(self)
        self._server_factory = server_factory
        self._server = None  # type: Optional[LocalDevServer]
        self.daemon = True

    def run(self):
        # type: () -> None
        self._server = self._server_factory()
        self._server.serve_forever()

    def shutdown(self):
        # type: () -> None
        if self._server is not None:
            self._server.shutdown()


class LocalChalice(Chalice):

    _THREAD_LOCAL = threading.local()

    # This is a known mypy bug where you can't override instance
    # variables with properties.  So this should be type safe, which
    # is why we're adding the type: ignore comments here.
    # See: https://github.com/python/mypy/issues/4125

    @property  # type: ignore
    def current_request(self):  # type: ignore
        # type: () -> Request
        return self._THREAD_LOCAL.current_request

    @current_request.setter
    def current_request(self, value):  # type: ignore
        # type: (Request) -> None
        self._THREAD_LOCAL.current_request = value


class CustomLocalChalice(LocalChalice):
    pass



import copy
import inspect

from typing import Any, List, Dict, Optional, Union  # noqa

from chalice.app import Chalice, RouteEntry, Authorizer, CORSConfig  # noqa
from chalice.app import ChaliceAuthorizer
from chalice.deploy.planner import StringFormat
from chalice.deploy.models import RestAPI  # noqa
from chalice.utils import to_cfn_resource_name


class SwaggerGenerator(object):

    _BASE_TEMPLATE = {
        'swagger': '2.0',
        'info': {
            'version': '1.0',
            'title': ''
        },
        'schemes': ['https'],
        'paths': {},
        'definitions': {
            'Empty': {
                'type': 'object',
                'title': 'Empty Schema',
            }
        }
    }  # type: Dict[str, Any]

    def __init__(self, region, deployed_resources):
        # type: (str, Dict[str, Any]) -> None
        self._region = region
        self._deployed_resources = deployed_resources

    def generate_swagger(self, app, rest_api=None):
        # type: (Chalice, Optional[RestAPI]) -> Dict[str, Any]
        api = copy.deepcopy(self._BASE_TEMPLATE)
        api['info']['title'] = app.app_name
        self._add_binary_types(api, app)
        self._add_route_paths(api, app)
        self._add_resource_policy(api, rest_api)
        self._add_vpc_endpoint(api, rest_api)
        return api

    def _add_resource_policy(self, api, rest_api):
        # type: (Dict[str, Any], Optional[RestAPI]) -> None
        if rest_api and rest_api.policy:
            api['x-amazon-apigateway-policy'] = rest_api.policy.document

    def _add_vpc_endpoint(self, api, rest_api):
        # type: (Dict[str, Any], Optional[RestAPI]) -> None
        if rest_api and rest_api.vpce_ids:
            api['x-amazon-apigateway-endpoint-configuration'] = {
                "vpcEndpointIds": rest_api.vpce_ids
            }

    def _add_binary_types(self, api, app):
        # type: (Dict[str, Any], Chalice) -> None
        api['x-amazon-apigateway-binary-media-types'] = app.api.binary_types

    def _add_route_paths(self, api, app):
        # type: (Dict[str, Any], Chalice) -> None
        for path, methods in app.routes.items():
            swagger_for_path = {}  # type: Dict[str, Any]
            api['paths'][path] = swagger_for_path

            cors_config = None
            methods_with_cors = []
            for http_method, view in methods.items():
                current = self._generate_route_method(view)
                if 'security' in current:
                    self._add_to_security_definition(
                        current['security'], api, view)
                swagger_for_path[http_method.lower()] = current
                if view.cors is not None:
                    cors_config = view.cors
                    methods_with_cors.append(http_method)

            # Chalice ensures that routes with multiple views have the same
            # CORS configuration. So if any entry has CORS enabled, use that
            # entry's CORS configuration for the preflight setup.
            if cors_config is not None:
                self._add_preflight_request(
                    cors_config, methods_with_cors, swagger_for_path)

    def _generate_security_from_auth_obj(self, api_config, authorizer):
        # type: (Dict[str, Any], Authorizer) -> None
        if isinstance(authorizer, ChaliceAuthorizer):
            auth_config = authorizer.config
            config = {
                'in': 'header',
                'type': 'apiKey',
                'name': auth_config.header,
                'x-amazon-apigateway-authtype': 'custom'
            }
            api_gateway_authorizer = {
                'type': 'token',
                'authorizerUri': self._auth_uri(authorizer)
            }
            if auth_config.execution_role is not None:
                api_gateway_authorizer['authorizerCredentials'] = \
                    auth_config.execution_role
            if auth_config.ttl_seconds is not None:
                api_gateway_authorizer['authorizerResultTtlInSeconds'] = \
                    auth_config.ttl_seconds
            config['x-amazon-apigateway-authorizer'] = api_gateway_authorizer
        else:
            config = authorizer.to_swagger()
        api_config.setdefault(
            'securityDefinitions', {})[authorizer.name] = config

    def _auth_uri(self, authorizer):
        # type: (ChaliceAuthorizer) -> str
        function_name = '%s-%s' % (
            self._deployed_resources['api_handler_name'],
            authorizer.config.name
        )
        return self._uri(
            self._deployed_resources['lambda_functions'][function_name]['arn'])

    def _add_to_security_definition(self, security,
                                    api_config, view):
        # type: (Any, Dict[str, Any], RouteEntry) -> None
        if view.authorizer is not None:
            self._generate_security_from_auth_obj(api_config, view.authorizer)
        for auth in security:
            name = list(auth.keys())[0]
            if name == 'api_key':
                # This is just the api_key_required=True config
                swagger_snippet = {
                    'type': 'apiKey',
                    'name': 'x-api-key',
                    'in': 'header',
                }  # type: Dict[str, Any]
                api_config.setdefault(
                    'securityDefinitions', {})[name] = swagger_snippet

    def _generate_route_method(self, view):
        # type: (RouteEntry) -> Dict[str, Any]
        current = {
            'consumes': view.content_types,
            'produces': ['application/json'],
            'responses': self._generate_precanned_responses(),
            'x-amazon-apigateway-integration': self._generate_apig_integ(
                view),
        }  # type: Dict[str, Any]
        docstring = inspect.getdoc(view.view_function)
        if docstring:
            doc_lines = docstring.splitlines()
            current['summary'] = doc_lines[0]
            if len(doc_lines) > 1:
                current['description'] = '\n'.join(doc_lines[1:]).strip('\n')
        if view.api_key_required:
            # When this happens we also have to add the relevant portions
            # to the security definitions.  We have to someone indicate
            # this because this neeeds to be added to the global config
            # file.
            current.setdefault('security', []).append({'api_key': []})
        if view.authorizer:
            current.setdefault('security', []).append(
                {view.authorizer.name: view.authorizer.scopes})
        if view.view_args:
            self._add_view_args(current, view.view_args)
        return current

    def _generate_precanned_responses(self):
        # type: () -> Dict[str, Any]
        responses = {
            '200': {
                'description': '200 response',
                'schema': {
                    '$ref': '#/definitions/Empty',
                }
            }
        }
        return responses

    def _uri(self, lambda_arn=None):
        # type: (Optional[str]) -> Any
        if lambda_arn is None:
            lambda_arn = self._deployed_resources['api_handler_arn']
        partition = lambda_arn.split(':')[1]
        return ('arn:{partition}:apigateway:{region}:lambda:path/2015-03-31'
                '/functions/{lambda_arn}/invocations').format(
                    partition=partition, region=self._region,
                    lambda_arn=lambda_arn)

    def _generate_apig_integ(self, view):
        # type: (RouteEntry) -> Dict[str, Any]
        apig_integ = {
            'responses': {
                'default': {
                    'statusCode': "200",
                }
            },
            'uri': self._uri(),
            'passthroughBehavior': 'when_no_match',
            'httpMethod': 'POST',
            'contentHandling': 'CONVERT_TO_TEXT',
            'type': 'aws_proxy',
        }
        return apig_integ

    def _add_view_args(self, single_method, view_args):
        # type: (Dict[str, Any], List[str]) -> None
        single_method['parameters'] = [
            {'name': name, 'in': 'path', 'required': True, 'type': 'string'}
            for name in view_args
        ]

    def _add_preflight_request(self, cors, methods, swagger_for_path):
        # type: (CORSConfig, List[str], Dict[str, Any]) -> None
        methods = methods + ['OPTIONS']
        allowed_methods = ','.join(methods)

        response_params = {
            'Access-Control-Allow-Methods': '%s' % allowed_methods
        }
        response_params.update(cors.get_access_control_headers())

        headers = {k: {'type': 'string'} for k, _ in response_params.items()}
        response_params = {'method.response.header.%s' % k: "'%s'" % v for k, v
                           in response_params.items()}

        options_request = {
            "consumes": ["application/json"],
            "produces": ["application/json"],
            "responses": {
                "200": {
                    "description": "200 response",
                    "schema": {"$ref": "#/definitions/Empty"},
                    "headers": headers
                }
            },
            "x-amazon-apigateway-integration": {
                "responses": {
                    "default": {
                        "statusCode": "200",
                        "responseParameters": response_params,
                    }
                },
                "requestTemplates": {
                    "application/json": "{\"statusCode\": 200}"
                },
                "passthroughBehavior": "when_no_match",
                "type": "mock",
                "contentHandling": "CONVERT_TO_TEXT"
            }
        }
        swagger_for_path['options'] = options_request


class CFNSwaggerGenerator(SwaggerGenerator):
    def __init__(self):
        # type: () -> None
        pass

    def _uri(self, lambda_arn=None):
        # type: (Optional[str]) -> Any
        return {
            'Fn::Sub': (
                'arn:${AWS::Partition}:apigateway:${AWS::Region}'
                ':lambda:path/2015-03-31'
                '/functions/${APIHandler.Arn}/invocations'
            )
        }

    def _auth_uri(self, authorizer):
        # type: (ChaliceAuthorizer) -> Any
        return {
            'Fn::Sub': (
                'arn:${AWS::Partition}:apigateway:${AWS::Region}'
                ':lambda:path/2015-03-31'
                '/functions/${%s.Arn}/invocations' % to_cfn_resource_name(
                    authorizer.name)
            )
        }


class TemplatedSwaggerGenerator(SwaggerGenerator):
    def __init__(self):
        # type: () -> None
        pass

    def _uri(self, lambda_arn=None):
        # type: (Optional[str]) -> Any
        return StringFormat(
            'arn:{partition}:apigateway:{region_name}:lambda:path/2015-03-31'
            '/functions/{api_handler_lambda_arn}/invocations',
            ['partition', 'region_name', 'api_handler_lambda_arn'],
        )

    def _auth_uri(self, authorizer):
        # type: (ChaliceAuthorizer) -> Any
        varname = '%s_lambda_arn' % authorizer.name
        return StringFormat(
            'arn:{partition}:apigateway:{region_name}:lambda:path/2015-03-31'
            '/functions/{%s}/invocations' % varname,
            ['partition', 'region_name', varname],
        )


class TerraformSwaggerGenerator(SwaggerGenerator):

    def __init__(self):
        # type: () -> None
        pass

    def _uri(self, lambda_arn=None):
        # type: (Optional[str]) -> Any
        return '${aws_lambda_function.api_handler.invoke_arn}'

    def _auth_uri(self, authorizer):
        # type: (ChaliceAuthorizer) -> Any
        return '${aws_lambda_function.%s.invoke_arn}' % (authorizer.name)



from chalice import Chalice

app = Chalice(app_name='{{app_name}}')


@app.lambda_function()
def first_function(event, context):
    return {'hello': 'world'}


@app.lambda_function()
def second_function(event, context):
    return {'hello': 'world2'}



import sys
import os
import json
import importlib
import logging
import functools

import click
from botocore.config import Config as BotocoreConfig
from botocore.session import Session
from typing import Any, Optional, Dict, MutableMapping, cast  # noqa

from chalice import __version__ as chalice_version
from chalice.awsclient import TypedAWSClient
from chalice.app import Chalice  # noqa
from chalice.config import Config
from chalice.config import DeployedResources  # noqa
from chalice.package import create_app_packager
from chalice.package import AppPackager  # noqa
from chalice.package import PackageOptions
from chalice.constants import DEFAULT_STAGE_NAME
from chalice.constants import DEFAULT_APIGATEWAY_STAGE_NAME
from chalice.constants import DEFAULT_ENDPOINT_TYPE
from chalice.logs import LogRetriever, LogEventGenerator
from chalice.logs import FollowLogEventGenerator
from chalice.logs import BaseLogEventGenerator
from chalice import local
from chalice.utils import UI  # noqa
from chalice.utils import PipeReader  # noqa
from chalice.deploy import deployer  # noqa
from chalice.deploy import validate
from chalice.invoke import LambdaInvokeHandler
from chalice.invoke import LambdaInvoker
from chalice.invoke import LambdaResponseFormatter


OptStr = Optional[str]
OptInt = Optional[int]


def create_botocore_session(profile=None, debug=False,
                            connection_timeout=None,
                            read_timeout=None,
                            max_retries=None):
    # type: (OptStr, bool, OptInt, OptInt, OptInt) -> Session
    s = Session(profile=profile)
    _add_chalice_user_agent(s)
    if debug:
        _inject_large_request_body_filter()
    config_args = {}  # type: Dict[str, Any]
    if connection_timeout is not None:
        config_args['connect_timeout'] = connection_timeout
    if read_timeout is not None:
        config_args['read_timeout'] = read_timeout
    if max_retries is not None:
        config_args['retries'] = {'max_attempts': max_retries}
    if config_args:
        config = BotocoreConfig(**config_args)
        s.set_default_client_config(config)
    return s


def _add_chalice_user_agent(session):
    # type: (Session) -> None
    suffix = '%s/%s' % (session.user_agent_name, session.user_agent_version)
    session.user_agent_name = 'aws-chalice'
    session.user_agent_version = chalice_version
    session.user_agent_extra = suffix


def _inject_large_request_body_filter():
    # type: () -> None
    log = logging.getLogger('botocore.endpoint')
    log.addFilter(LargeRequestBodyFilter())


class NoSuchFunctionError(Exception):
    """The specified function could not be found."""

    def __init__(self, name):
        # type: (str) -> None
        self.name = name
        super(NoSuchFunctionError, self).__init__()


class UnknownConfigFileVersion(Exception):
    def __init__(self, version):
        # type: (str) -> None
        super(UnknownConfigFileVersion, self).__init__(
            "Unknown version '%s' in config.json" % version)


class LargeRequestBodyFilter(logging.Filter):
    def filter(self, record):
        # type: (Any) -> bool
        # Note: the proper type should be "logging.LogRecord", but
        # the typechecker complains about 'Invalid index type "int" for "dict"'
        # so we're using Any for now.
        if record.msg.startswith('Making request'):
            if record.args[0].name in ['UpdateFunctionCode', 'CreateFunction']:
                # When using the ZipFile argument (which is used in chalice),
                # the entire deployment package zip is sent as a base64 encoded
                # string.  We don't want this to clutter the debug logs
                # so we don't log the request body for lambda operations
                # that have the ZipFile arg.
                record.args = (record.args[:-1] +
                               ('(... omitted from logs due to size ...)',))
        return True


class CLIFactory(object):
    def __init__(self, project_dir, debug=False, profile=None, environ=None):
        # type: (str, bool, Optional[str], Optional[MutableMapping]) -> None
        self.project_dir = project_dir
        self.debug = debug
        self.profile = profile
        if environ is None:
            environ = dict(os.environ)
        self._environ = environ

    def create_botocore_session(self, connection_timeout=None,
                                read_timeout=None, max_retries=None):
        # type: (OptInt, OptInt, OptInt) -> Session
        return create_botocore_session(profile=self.profile,
                                       debug=self.debug,
                                       connection_timeout=connection_timeout,
                                       read_timeout=read_timeout,
                                       max_retries=max_retries)

    def create_default_deployer(self, session, config, ui):
        # type: (Session, Config, UI) -> deployer.Deployer
        return deployer.create_default_deployer(session, config, ui)

    def create_plan_only_deployer(self, session, config, ui):
        # type: (Session, Config, UI) -> deployer.Deployer
        return deployer.create_plan_only_deployer(session, config, ui)

    def create_deletion_deployer(self, session, ui):
        # type: (Session, UI) -> deployer.Deployer
        return deployer.create_deletion_deployer(
            TypedAWSClient(session), ui)

    def create_deployment_reporter(self, ui):
        # type: (UI) -> deployer.DeploymentReporter
        return deployer.DeploymentReporter(ui=ui)

    def create_config_obj(self, chalice_stage_name=DEFAULT_STAGE_NAME,
                          autogen_policy=None,
                          api_gateway_stage=None,
                          user_provided_params=None):
        # type: (str, Optional[bool], str, Optional[Dict[str, Any]]) -> Config
        if user_provided_params is None:
            user_provided_params = {}
        default_params = {'project_dir': self.project_dir,
                          'api_gateway_stage': DEFAULT_APIGATEWAY_STAGE_NAME,
                          'api_gateway_endpoint_type': DEFAULT_ENDPOINT_TYPE,
                          'autogen_policy': True}
        try:
            config_from_disk = self.load_project_config()
        except (OSError, IOError):
            raise RuntimeError("Unable to load the project config file. "
                               "Are you sure this is a chalice project?")
        except ValueError as err:
            raise RuntimeError("Unable to load the project config file: %s"
                               % err)

        self._validate_config_from_disk(config_from_disk)
        if autogen_policy is not None:
            user_provided_params['autogen_policy'] = autogen_policy
        if self.profile is not None:
            user_provided_params['profile'] = self.profile
        if api_gateway_stage is not None:
            user_provided_params['api_gateway_stage'] = api_gateway_stage
        config = Config(chalice_stage=chalice_stage_name,
                        user_provided_params=user_provided_params,
                        config_from_disk=config_from_disk,
                        default_params=default_params)
        user_provided_params['chalice_app'] = functools.partial(
            self.load_chalice_app, config.environment_variables)
        return config

    def _validate_config_from_disk(self, config):
        # type: (Dict[str, Any]) -> None
        string_version = config.get('version', '1.0')
        try:
            version = float(string_version)
            if version > 2.0:
                raise UnknownConfigFileVersion(string_version)
        except ValueError:
            raise UnknownConfigFileVersion(string_version)

    def create_app_packager(self, config, options, package_format,
                            template_format, merge_template=None):
        # type: (Config, PackageOptions, str, str, OptStr) -> AppPackager
        return create_app_packager(
            config, options, package_format, template_format,
            merge_template=merge_template)

    def create_log_retriever(self, session, lambda_arn, follow_logs):
        # type: (Session, str, bool) -> LogRetriever
        client = TypedAWSClient(session)
        if follow_logs:
            event_generator = cast(BaseLogEventGenerator,
                                   FollowLogEventGenerator(client))
        else:
            event_generator = cast(BaseLogEventGenerator,
                                   LogEventGenerator(client))
        retriever = LogRetriever.create_from_lambda_arn(event_generator,
                                                        lambda_arn)
        return retriever

    def create_stdin_reader(self):
        # type: () -> PipeReader
        stream = click.get_binary_stream('stdin')
        reader = PipeReader(stream)
        return reader

    def create_lambda_invoke_handler(self, name, stage):
        # type: (str, str) -> LambdaInvokeHandler
        config = self.create_config_obj(stage)
        deployed = config.deployed_resources(stage)
        try:
            resource = deployed.resource_values(name)
            arn = resource['lambda_arn']
        except (KeyError, ValueError):
            raise NoSuchFunctionError(name)

        function_scoped_config = config.scope(stage, name)
        # The session for max retries needs to be set to 0 for invoking a
        # lambda function because in the case of a timeout or other retriable
        # error the underlying client will call the function again.
        session = self.create_botocore_session(
            read_timeout=function_scoped_config.lambda_timeout,
            max_retries=0,
        )

        client = TypedAWSClient(session)
        invoker = LambdaInvoker(arn, client)

        handler = LambdaInvokeHandler(
            invoker,
            LambdaResponseFormatter(),
            UI(),
        )

        return handler

    def load_chalice_app(self, environment_variables=None,
                         validate_feature_flags=True):
        # type: (Optional[MutableMapping], Optional[bool]) -> Chalice
        # validate_features indicates that we should validate that
        # any expiremental features used have the appropriate feature flags.
        if self.project_dir not in sys.path:
            sys.path.insert(0, self.project_dir)
        # The vendor directory has its contents copied up to the top level of
        # the deployment package. This means that imports will work in the
        # lambda function as if the vendor directory is on the python path.
        # For loading the config locally we must add the vendor directory to
        # the path so it will be treated the same as if it were running on
        # lambda.
        vendor_dir = os.path.join(self.project_dir, 'vendor')
        if os.path.isdir(vendor_dir) and vendor_dir not in sys.path:
            # This is a tradeoff we have to make for local use.
            # The common use case of vendor/ is to include
            # extension modules built for AWS Lambda.  If you're
            # running on a non-linux dev machine, then attempting
            # to import these files will raise exceptions.  As
            # a workaround, the vendor is added to the end of
            # sys.path so it's after `./lib/site-packages`.
            # This gives you a change to install the correct
            # version locally and still keep the lambda
            # specific one in vendor/
            sys.path.append(vendor_dir)
        if environment_variables is not None:
            self._environ.update(environment_variables)
        try:
            app = importlib.import_module('app')
            chalice_app = getattr(app, 'app')
        except SyntaxError as e:
            message = (
                'Unable to import your app.py file:\n\n'
                'File "%s", line %s\n'
                '  %s\n'
                'SyntaxError: %s'
            ) % (getattr(e, 'filename'), e.lineno, e.text, e.msg)
            raise RuntimeError(message)
        if validate_feature_flags:
            validate.validate_feature_flags(chalice_app)
        return chalice_app

    def load_project_config(self):
        # type: () -> Dict[str, Any]
        """Load the chalice config file from the project directory.

        :raise: OSError/IOError if unable to load the config file.

        """
        config_file = os.path.join(self.project_dir, '.chalice', 'config.json')
        with open(config_file) as f:
            return json.loads(f.read())

    def create_local_server(self, app_obj, config, host, port):
        # type: (Chalice, Config, str, int) -> local.LocalDevServer
        return local.create_local_server(app_obj, config, host, port)

    def create_package_options(self):
        # type: () -> PackageOptions
        """Create the package options that are required to target regions."""
        s = Session(profile=self.profile)
        client = TypedAWSClient(session=s)
        return PackageOptions(client)
